#!/usr/bin/env python
from bidali import LSD
import gzip, tarfile, biomart, pandas as pd
from io import TextIOWrapper, StringIO
from .ensembl import get_biomart

def get_proteinHalfLifes(datadir=LSD.datadir+'PMID_published_tables/21593866_protein_half_lifes/'):
    """
    Mouse protein half lifes

    >>> phl = get_proteinHalfLifes()
    """
    phl = pd.read_excel(datadir+"nature10098-s5.xls",sheetname=0)
    return phl

def get_cyclebase(datadir=LSD.datadir+'PMID_published_tables/25378319_cyclebase/'):
    """
    Info: https://cyclebase.org/
    Info: http://genome-www.stanford.edu/Human-CellCycle/HeLa/data/dataPlusScores_all5.txt

    cyb = get_cyclebase()
    """
    tf = tarfile.open(datadir+'human_timecourses.tar')
    #metadata = pd.read_table(TextIOWrapper(tf.extractfile('human_metadata.tsv')))
    hexp = pd.read_table(TextIOWrapper(tf.extractfile('human_experiments.tsv')))
    #whitfieldData = pd.read_table(datadir+'dataPlusScores_all5.txt',index_col='UID')
    #whitfieldData['gene_name'] = whitfieldData.NAME.fillna('').apply(lambda x: x.split()[1] if len(x.split())>1 else '')
    #hexp['gene_name'] = hexp.identifier.apply(lambda x: whitfieldData.ix[x].gene_name)
    mapping = pd.read_table(datadir+'cyb_probeset_mapping.txt',header=None,names=('species','protein','code_id'))

    # Periodic genes
    tf = tarfile.open(datadir+'human_periodic.tar')
    periodicGenes = pd.read_table(TextIOWrapper(tf.extractfile('human_periodic.tsv')))
    bmrt = get_biomart()
    protgenenames = bmrt[['Gene name','Protein stable ID']].dropna().set_index('Protein stable ID')
    periodicGenes = periodicGenes.join(other=protgenenames,on='gene').set_index('Gene name')

    return LSD.Dataset(
        periodicGenes=periodicGenes,
        hexp=hexp,
        mapping=mapping
    )



#!/usr/bin/env python3
#R exposed top routines
import pickle, re, sys, operator
from bidali import LSD
import pandas as pd, numpy as np
from collections import OrderedDict
import matplotlib.pyplot as plt
from rpy2.rinterface import RRuntimeError
from rpy2.robjects.packages import importr
import rpy2.robjects as ro
#Activate automatic pandas/r conversion
from rpy2.robjects import pandas2ri
pandas2ri.activate()

# General importr's
base = importr('base')
stats = importr('stats')
utils = importr('utils')

# Differential expression analysis
def prepareDesign(metadata,design,reflevels=None,RReturnOnly=True):
    """
    Takes pandas metadata table and design string
    to return R design object

    e.g. design = '~0+treatment+batch'
    """
    metacols_r = {
        col: ro.r.relevel(
            ro.r.factor(metadata[col]),
            ref=reflevels[col] if reflevels else None
        )
        for col in metadata
    }
    metadata_r = base.data_frame(**metacols_r)
    design_r = stats.model_matrix(ro.r.formula(design),data=metadata_r)
    design_r.colnames = ro.StrVector(
        [c.replace(':','.') for c in ro.r.colnames(design_r)]
    ) # resolves naming issue when using interaction factors in design
    #design = pd.DataFrame(np.array(design_r),columns=design_r.colnames,index=design_r.rownames)
    #design.doxnox*2 + (-design.shairpinTBX2sh25 - design.shairpinTBX2sh27)*3*design.doxnox
    if RReturnOnly:
        return design_r
    else:
        design_p = pd.DataFrame(pandas2ri.ri2py(design_r), index = metadata.index, columns = design_r.colnames)
        return design_r, design_p
    
def prepareContrasts(design_r,contrasts, RReturnOnly = True):
    """
    Expects R design_r and list of contrasts that contain design_r column names
    Returns contrasts_r and also pd.DataFrame of contrasts_r if not RReturnOnly

    e.g. contrasts = [
      'treatmentSHC002_dox - treatmentTBX2sh25_dox','treatmentSHC002_dox - treatmentTBX2sh27_dox',
      'treatmentSHC002_nox - treatmentTBX2sh25_nox','treatmentSHC002_nox - treatmentTBX2sh27_nox'
    ]
    """
    # import R limma package
    limma = importr('limma')
    contrasts_r = limma.makeContrasts(*contrasts,levels=design_r)
    if RReturnOnly:
        return contrasts_r
    else:
        contrasts_p = pd.DataFrame(
            pandas2ri.ri2py(contrasts_r),
            columns=ro.r.colnames(contrasts_r),
            index=ro.r.rownames(contrasts_r)
        )
        return contrasts_r, contrasts_p

def DEA(counts,design_r,contrasts=None,adjust_method='BH'):
    """
    contrasts needs to be a dictionary with string keys and values to be used as design contrasts or 
    a list of int values to be used as design coefficients

    Returns results and pd.DataFrame of normalized counts
    """
    if isinstance(contrasts, list):
        coefs = True
        if not all(isinstance(k, int) for k in contrasts):
            raise ValueError('coefficient list should be all integers')
        coefnames = list(ro.r.colnames(design_r))
        contrasts = OrderedDict([(coefnames[c-1],c) for c in contrasts])
    else:
        coefs = False
        if not (all(isinstance(k, str) for k in contrasts) and
                all(isinstance(v, str) for v in contrasts.values())):
            raise ValueError('contrast dict should be all string pairs')
    
    # import R limma package
    limma = importr('limma')
    # tranform counts with voom
    voomedCounts_r = limma.voom(counts,design=design_r,plot=True,normalize="quantile")
    fit_r = limma.lmFit(voomedCounts_r,design_r)
    fit_r = limma.eBayes(fit_r)
    coefficients_r = fit_r.rx2('coefficients') #fit_r$coefficients
    if coefs:
        fit_contrasts_r = fit_r
    else:
        contrasts_r, contrasts_p = prepareContrasts(design_r, contrasts.values(), RReturnOnly = False)
        fit_contrasts_r = limma.contrasts_fit(fit_r, contrasts_r)
        fit_contrasts_r = limma.eBayes(fit_contrasts_r)
    print(ro.r.summary(fit_contrasts_r))

    #Full results
    results = OrderedDict()
    for res in contrasts:
        result_r = limma.topTable(
            fit_contrasts_r,coef=contrasts[res],n=len(counts),adjust_method=adjust_method
        )
        results[res] = pandas2ri.ri2py(result_r)
        results[res].index = ro.r.rownames(result_r)
        #results[res]['gene_label'] = results[res].index.map(lambda x: counts.index[int(x)-1])
        print('# sig',res,'->',(results[res]['adj.P.Val']<=0.05).sum())
        
    return results, pd.DataFrame(
        pandas2ri.ri2py(voomedCounts_r.rx2('E')),
        columns=counts.columns,index=counts.index
    )
#limma volcanoplot and plotMDS need to be done either in qconsole or direct R environment

# Gene-set enrichment analysis
## limma based
def get_gcindices_r(countsGeneLabels,correctBackground=False,remove_empty=True):
    """
    >>> indices_r = get_gcindices(counts.index,correctBackground=False) # doctest: +SKIP

    TODO overlapping with genesets2indices_r => refactor code
    """
    limma = importr('limma')
    gc = LSD.get_msigdb6()

    if correctBackground:
        gc = {gsc:{gs:[g for g in gc[gsc][gs] if g in countsGeneLabels] for gs in gc[gsc]} for gsc in gc}
        
    countsGeneLabels_r = ro.StrVector(countsGeneLabels)        
    gc_indices_r = {gsc:limma.ids2indices(ro.ListVector(gc[gsc]),countsGeneLabels_r,remove_empty=remove_empty)
                    for gsc in gc}
    
    return gc_indices_r

def romer(counts_r,index_r,design_r,contrast_r):
    """
    counts_r -> should be log-expression values (voom transformed RNAseq counts also good)
    index_r -> collection of geneses (ListVector of IntVector's)
    design_r -> design matrix
    contrast_r -> coef or contrast vector
      e.g. contrasts_r.rx(True,"celllineIMR32.shairpinsh25+celllineIMR32.shairpinsh27"

    >>> romer(counts_r,index_r,design_r,contrast_r) # doctest: +SKIP
    """
    limma = importr('limma')
    rr = limma.romer(counts_r,index=index_r,design=design_r,contrast=contrast_r)
    rr_py = pd.DataFrame(pandas2ri.ri2py(rr),columns=base.colnames(rr),index=base.rownames(rr))
    rr_py['Up_adj'] = stats.p_adjust(rr.rx(True,'Up'),method='fdr')
    rr_py['Down_adj'] = stats.p_adjust(rr.rx(True,'Down'),method='fdr')
    rr_py['min_p_adj'] = rr_py[['Up_adj','Down_adj']].T.min()
    rr_py.sort_values('min_p_adj',inplace=True)
    return rr_py

def romer_fullMSigDB(counts_r,countsGeneLabels,**kwargs):
    """
    >>> romer_fullMSigDB(counts_r,countsGeneLabels) # doctest: +SKIP
    """
    gc = get_gcindices_r(countsGeneLabels)
    return OrderedDict([
        (gsc,romer(counts_r,index_r=gc[gsc],**kwargs))
        for gsc in sorted(gc)
    ])

def mroast(counts_r,index_r,design_r,contrast_r,set_statistic="mean"):
    """
    >>> mroast(counts_r,index_r,design_r,contrast_r) # doctest: +SKIP
    """
    limma = importr('limma')
    return base.data_frame(limma.mroast(counts_r,index_r,design=design_r,contrast=contrast_r,set_statistic=set_statistic))

def genesets2indices_r(genesets, geneLabels, remove_empty=True):
    """
    genesets should be a dictionary of geneset lists, and geneLabels a list of gene labels
    >>> from .tests.test_retro import testGenesets, testCountsGenelabels
    >>> print(genesets2indices_r(testGenesets,testCountsGenelabels))
    [[1]]
    [1] 1 2
    <BLANKLINE>
    [[2]]
    [1] 2 3
    <BLANKLINE>
    <BLANKLINE>

    TODO test function for gene level conversion between python and R
    """
    limma = importr('limma')
    genesets_r = ro.ListVector({gs:ro.StrVector(genesets[gs]) for gs in genesets})
    return limma.ids2indices(genesets_r, geneLabels, remove_empty = remove_empty)

def barcodeplot(fit_r,contrast,indices,geneset,geneset2=None,pngname=None,width=512,height=512,**kwargs):
    """
    >>> barcodeplot(fit_r,3,indices['C5'],'GO_GLUTATHIONE_TRANSFERASE_ACTIVITY',pngname=f.name) # doctest: +SKIP
    """
    limma = importr('limma')
    if pngname:
        grdevices = importr('grDevices')
        grdevices.png(file=pngname, width=width, height=height)
    limma.barcodeplot(fit_r.rx2('t').rx(True,contrast),indices.rx2(geneset),
                      index2 = indices.rx2(geneset2) if geneset2 else ro.NULL,**kwargs)
    if pngname: grdevices.dev_off()

## fgsea based
def fgsea(genesets_r, ranks, minSize = 15, maxSize = 500, nperm = 10000):
    """
    Ranks should be pd.Series with as index all genes in geneLabels used for generating genesets_r
    with e.g. genesets2indices_r

    TODO test function for gene level conversion between python and R
    """
    fgsea = importr('fgsea')
    fgseaRes_r = fgsea.fgsea(
        pathways = genesets_r, stats = pd.Series(list(ranks),index=range(1,len(ranks)+1)),
        minSize = minSize, maxSize = maxSize, nperm = nperm
    )
    fgseaRes = pandas2ri.ri2py(fgseaRes_r.rx(True,-8))
    fgseaRes['leadingEdge'] = [
        [ranks.index[int(gn)-1] for gn in fgseaRes_r.rx2(i,'leadingEdge')]
        for i in range(1,len(fgseaRes_r.rownames)+1)
    ]
    return fgseaRes.sort_values('padj')

## rank sums
def rankSumProbsMSigDB(ranks,universe,adjpmethod='fdr'):
    from bidali.fegnome import RankSumResult,rankSumProbability
    gc = LSD.get_msigdb6()
    results = OrderedDict()
    for gsc in sorted(gc):
        results[gsc] = pd.DataFrame({gs:rankSumProbability(ranks,gc[gsc][gs]) for gs in gc[gsc]}).T
        results[gsc].columns = RankSumResult._fields
        results[gsc]['fepa'] = stats.p_adjust(results[gsc].fe_p,method=adjpmethod)
        results[gsc]['respa'] = stats.p_adjust(results[gsc].probability,method=adjpmethod)
        results[gsc].sort_values(['fepa','respa'],inplace=True)

    return results



# -*- coding: utf-8 -*-
"""Format types for localised data structures.
"""
from bidali import LSD
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, zipfile, pandas as pd

class XML(object):
    """Class to work with xml files.
    
    Mainly methods to access specific data easier, than directry to Python standard lib.

    Args:
        xmlfile (str): Can either be a filename or an already parsed tree (but not the root element).
    """
    def __init__(self,xmlfile,namespace_symbol=None,namespace_url=None,schema=None):
        import xml.etree.ElementTree as ET
        if isintance(xmlfile,str):
            with open(xmlfile) as f:
                self.tree = ET.parse(f)
        else:
            self.tree = xmlfile
        self.root = self.tree.getroot()
        self.namespace_symbol = namespace_symbol
        if namespace_symbol:
            self.tree_namespace = {self.namespace_symbol:namespace_url}
        self.schema = schema

    def __getitem__(self,xpath):
        #prepending namespace to all path elements
        if self.namespace_symbol: 
            xpath = '/'.join(
                ':'.join((self.namespace_symbol,d)) for d in xpath.split('/')
            )
        return self.root.findall(xpath, self.tree_namespace)

    def get_tags(self):
        return {
            e.tag.replace(
                '{'+self.tree_namespace[self.namespace_symbol]+'}',
                ''
            ) if self.namespace_symbol else e
            for e in self.tree.iter()
        }

    def validate(self):
        import xmlschema
        xs = xmlschema.XMLSchema(self.schema, converter=xmlschema.BadgerFishConverter)
        xs.validate(self.tree)



# -*- coding: utf-8 -*-
"""HGNC genenames

Reference: https://www.genenames.org/
"""
from urllib import error,parse
from urllib.request import Request,urlopen
import json

# Get the whole table
from bidali.LSD.dealer.ensembl import get_genenames

# genenames API
def fetchGenenamesJSON(symbol,alias=False):
    """
    Requires a gene symbol
    """
    if not alias:
        requesturl = 'http://rest.genenames.org/fetch/symbol/{}'
    else:
        requesturl = 'http://rest.genenames.org/fetch/alias_symbol/{}'
    requesturl = requesturl.format(parse.quote(symbol))
    request = Request(requesturl,
                      headers={'Accept':'application/json'})
    response = urlopen(request)
    content = response.read()
    if content:
        return json.loads(content.decode())
    
def fetchAliases(symbol,unknown_action='raise'):
    """
    Returns list of all aliases for a symbol.
    First alias in list is official symbol.

    Args:
        symbol (str): Symbol name.
        unknown_action (option):
          'raise': raise KeyError
          'list': [symbol]
          'none': None
    """
    symbolJSON = fetchGenenamesJSON(symbol)
    if not symbolJSON['response']['numFound']:
        symbolJSON = fetchGenenamesJSON(symbol,alias=True)
    try:
        symbols = [symbolJSON['response']['docs'][0]['symbol']
        ]+symbolJSON['response']['docs'][0]['alias_symbol']
        return symbols
    except KeyError:
        #Not a recognised symbol
        if unknown_action == 'raise': raise
        elif unknown_action == 'list': return [symbol]
        else: return None



#!/usr/bin/env python
from bidali import LSD
import pandas as pd, numpy as np, gzip
from os.path import expanduser, exists
from itertools import count
from bidali.LSD import storeDatasetLocally, datadir, Dataset

#TODO => NRC segmentation data from ~/Dropbiz/Lab/z_archive/AFW/project CONEXIC/NRC data/01. JISTIC

@storeDatasetLocally
def get_NRC(datadir=datadir+'NRC_data_AFW/'):
    # Metadata
    metadata = pd.read_excel(datadir+'20111216_NRC_samples.xlsx',skiprows=4)
    metadata.pop('Unnamed: 0')
    metadata.index = metadata.pop('NRC_ID')

    # Expression data
    exprdata = pd.read_table(datadir+'mRNA_expression_data.txt',index_col='Probeset')
    probeinfo = pd.read_table(datadir+'mRNA_info.txt',index_col='Probeset').dropna()
    print('Only retaining exprdata for HUGO genes')
    exprdata = exprdata[exprdata.index.isin(probeinfo.index)]
    probeinfo = pd.DataFrame([(h,v) for v in exprdata.index for h in probeinfo.ix[v]['HUGO'].split(',')],
                             columns=['HUGO','Probeset'])
    print('For genes with multiple probes, taking median')
    exprdata = probeinfo.join(other=exprdata,on='Probeset')
    del exprdata['Probeset']
    exprdata = exprdata.groupby('HUGO').median()
    print('Removing empty stringed gene name. Genes with multiple probes -> median taken.')
    exprdata.drop('',inplace=True)

    # Copy number variant data
    lo=LSD.get_liftover(frm=18,to=38)
    segments = pd.read_table(expanduser('~/Dropbox (speleman lab)/\
Lab/z_archive/AFW/project_CONEXIC/NRC_data/01_JISTIC/20111106_NRCdata_wavecorrected_en_Dublin_CBS.txt'))
    segments.chromosome = segments.chromosome.apply(
        lambda x: 'chrX' if x == 23 else 'chrY' if x == 24 else 'chr{}'.format(x))
    segments['Start38'] = segments.T.apply(
        lambda x: lo.convert_coordinate(x.chromosome,x.chr_start)).apply(lambda x: x[0][1] if x else np.nan
    )
    segments['End38'] = segments.T.apply(
        lambda x: lo.convert_coordinate(x.chromosome,x.chr_end)).apply(lambda x: x[0][1] if x else np.nan
    )
    del lo, segments['chr_start'], segments['chr_end']
    print(segments.shape[0],'before converting to hg38 from hg18')
    segments = segments.dropna().copy()
    print(segments.shape[0],'after converting to hg38')

    ## Assign genes to regions
    genannot = LSD.get_ensemblGeneannot()
    # def get_genes(region,context):
    #     """
    #     segments dataframe needs to be sorted according to alphabetical chrom name, and numerical location
    #     """
    #     #Reset gene_iter if different context
    #     if context != get_genes.previousContext:
    #         get_genes.genes_iter = genannot.all_features(featuretype='gene',order_by=('seqid','start'))
    #         get_genes.previousRunGene = False
    #         get_genes.previousContext = context
        
    #     genes = []
    #     if (get_genes.previousRunGene and
    #         region.chromosome[3:] == get_genes.previousRunGene.chrom and
    #         region.Start38 <= get_genes.previousRunGene.start and
    #         region.End38 > get_genes.previousRunGene.start):
    #         genes.append(get_genes.previousRunGene.attributes['gene_name'][0])
    #         get_genes.previousRunGene = False
            
    #     starting = True
    #     for gene in get_genes.genes_iter:
    #         if starting:
    #             if (
    #                     (region.chromosome[3:] > gene.chrom) or
    #                     (region.Start38 > gene.start)
    #             ):
    #                 continue
    #             starting = False
    #         if (region.chromosome[3:] == gene.chrom and region.Start38 <= gene.start and region.End38 > gene.start):
    #             genes.append(gene.attributes['gene_name'][0])
    #         else:
    #             get_genes.previousRunGene = gene
    #             break
    #     return genes
    # get_genes.previousContext = None

    # segments.sort_values(['exp_id','chromosome','Start38'],inplace=True)
    # segments['genes'] = segments.T.apply(lambda x: get_genes(x,context=x.exp_id))
    
    segments['genes'] = segments.T.apply(
        lambda x: {f.attributes['gene_name'][0] for f in genannot.region('{}:{}-{}'.format(
            x.chromosome[3:],int(x.Start38),int(x.End38)),featuretype='gene')}
    )
    segments['nrGenes'] = segments.genes.apply(len)
    del genannot#, get_genes
    #To set cut offs look at hist => aCGH.log2ratio.hist(ax=ax,bins='auto')
    segments['annotation'] = segments.ratio.apply(lambda x: 'gain' if x > 0.3 else ('loss' if x < -0.3 else 'normal'))

    geneCNVannotation = {}
    for sname,sample in segments.groupby('exp_id'):
        geneCNVannotation[sname] = {}
        for i in sample.T:
            for g in sample.ix[i].genes: geneCNVannotation[sname][g] = sample.ix[i].annotation
    geneCNVannotation = pd.DataFrame(geneCNVannotation)
    geneCNVannotation = geneCNVannotation.fillna('normal')

    #aCGH = pd.read_table(datadir+'20170214_untransformedR2_aCGH.txt',skiprows=44,
    #                     names=open(datadir+'20170214_untransformedR2_aCGH.txt').readline().split())
    #aCGH.pop('probeset')
    #aCGH.index = aCGH.pop('#H:hugo')

    return Dataset(**locals())

@storeDatasetLocally
def get_FischerData():
    filteredOn = { #For reference, to know how dataset has been filtered
        'minimalSurvivalLastFollowup': 365*5
        }
    
    #Metadata
    metadata = pd.read_table(gzip.open(datadir+
                                       "R2_grabbed_data/Fischer498/metadata_src/GSE49710_series_matrix.txt.gz",'rt',
                                       encoding="UTF-8"),
                             skiprows=47,skipfooter=44799-66,engine='python',header=None)
    metadata.index = metadata[0].apply(lambda x: x.replace('!',''))
    del metadata[0]
    metadata = metadata.T
    i = count()
    metadata.columns = [c.replace('ch1',str(next(i))) if c.startswith('Sample_char') else c for c in metadata.columns]
    del i, metadata['Sample_source_name_ch1'], metadata['Sample_status'], metadata['Sample_organism_ch1']
    metadata.columns = [metadata[c][metadata.first_valid_index()].split(':')[0].replace(' ','_')
                        if c.startswith('Sample_char') else c for c in metadata.columns]
    metadata = metadata.applymap(lambda x: x.split(': ')[1] if ': ' in x else x)
    
    metadatasurv = pd.read_table(gzip.open(datadir+"R2_grabbed_data/Fischer498/metadata_src/GSE62564_series_matrix.txt.gz",'rt',encoding="UTF-8"),
                                 skiprows=51,skipfooter=102-74,engine='python')
    metadatasurv.index = metadatasurv['!Sample_geo_accession'].apply(lambda x: x.replace('!',''))
    del metadatasurv['!Sample_geo_accession']
    metadatasurv = metadatasurv.T
    metadatasurv.columns = range(len(metadatasurv.columns))
    metadatasurv = metadatasurv[list(range(7,21))]
    metadatasurv.columns = [v.split(':')[0].replace(' ','_') for v in metadatasurv.ix[metadatasurv.first_valid_index()]]
    metadatasurv = metadatasurv.applymap(lambda x: x.split(': ')[1] if not x is np.nan else x)
    metadatasurv.index = metadata.index #Both sample sets are sorted the same way, but different gse names
    assert sum(metadatasurv.age == metadata.age_at_diagnosis) == len(metadata)
    metadata['overall_survival'] = metadatasurv.os_day.apply(int)
    metadata['eventfree_survival'] = metadatasurv.efs_day.apply(int)
    del metadatasurv
    metadata.index = metadata.Sample_geo_accession.apply(lambda x: x.lower())
    metadata.Sample_title = metadata.Sample_title.apply(lambda x: x[5:].replace(' patient ',''))
    metadata.death_from_disease = metadata.death_from_disease == '1'
    metadata.progression = metadata.progression == '1'

    #Expression data
    exprdata = pd.read_table(datadir+'R2_grabbed_data/Fischer498/GSE49710_R2.txt')
    exprdata.index = exprdata.pop('#H:hugo')
    del exprdata['probeset']

    #aCGH
    aCGH = pd.read_table(datadir+'R2_grabbed_data/Fischer498/SEQC_aCGH/SEQC_aCGH_all_146.txt')
    geosearch = metadata[['Sample_title','Sample_geo_accession']].copy()
    geosearch.Sample_geo_accession = geosearch.index
    geosearch.index = geosearch.Sample_title
    aCGH.Sample = aCGH.Sample.apply(lambda x: geosearch.ix[x].Sample_geo_accession)
    del geosearch
    aCGH['log2ratio'] = (aCGH.CN/2).apply(np.log2)
    #Convert coordinates to hg38
    lo = LSD.get_lift19to38()
    aCGH['Start38'] = aCGH.T.apply(lambda x: lo.convert_coordinate(x.Chromosome,x.Start)).apply(lambda x: x[0][1] if x else np.nan)
    aCGH['End38'] = aCGH.T.apply(lambda x: lo.convert_coordinate(x.Chromosome,x.End)).apply(lambda x: x[0][1] if x else np.nan)
    del lo, aCGH['Start'], aCGH['End']
    aCGH = aCGH.dropna().copy()
    #Assign genes to regions
    genannot = LSD.get_ensemblGeneannot()
    aCGH['genes'] = aCGH.T.apply(lambda x: {f.attributes['gene_name'][0] for f in genannot.region('{}:{}-{}'
                                                    .format(x.Chromosome[3:],int(x.Start38),int(x.End38)),featuretype='gene')})
    aCGH['nrGenes'] = aCGH.genes.apply(len)
    del genannot
    #To set cut offs look at hist => aCGH.log2ratio.hist(ax=ax,bins='auto')
    aCGH['annotation'] = aCGH.log2ratio.apply(lambda x: 'gain' if x > 0.3 else ('loss' if x < -0.3 else 'normal'))

    # Filter patients whom according to metadata survived, but had last follow up before 5 years
    metadata = metadata[metadata.death_from_disease |
        (~metadata.death_from_disease &
        (metadata.overall_survival > filteredOn['minimalSurvivalLastFollowup']))]
    exprdata = exprdata[metadata.index]
    aCGH = aCGH[aCGH.Sample.isin(metadata.index)]
    
    return Dataset(**locals())

@storeDatasetLocally
def get_SequencedFischerData():
    """
    For the sequence data, the Trente TUC files were chosen, following
    'cufflinks on RefSeq genes and reported as log2(1 + FPKM)' see TUC readme

    Reference: https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE49711

    Source: https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE49711&format=file&file=GSE49711%5FSEQC%5FNB%5FTUC%5FG%5Flog2%2Etxt%2Egz
    """
    filteredOn = { #For reference, to know how dataset has been filtered
        'minimalSurvivalLastFollowup': 365*5
        }
    print('Filtering settings:', filteredOn)

    #Metadata
    metadata = pd.read_table(gzip.open(datadir+
                                       "R2_grabbed_data/Fischer498/metadata_src/GSE49710_series_matrix.txt.gz",'rt',
                                       encoding="UTF-8"),
                             skiprows=47,skipfooter=44799-66,engine='python',header=None)
    metadata.index = metadata[0].apply(lambda x: x.replace('!',''))
    del metadata[0]
    metadata = metadata.T
    i = count()
    metadata.columns = [c.replace('ch1',str(next(i))) if c.startswith('Sample_char') else c for c in metadata.columns]
    del i, metadata['Sample_source_name_ch1'], metadata['Sample_status'], metadata['Sample_organism_ch1']
    metadata.columns = [metadata[c][metadata.first_valid_index()].split(':')[0].replace(' ','_')
                        if c.startswith('Sample_char') else c for c in metadata.columns]
    metadata = metadata.applymap(lambda x: x.split(': ')[1] if ': ' in x else x)
    
    metadatasurv = pd.read_table(gzip.open(datadir+"R2_grabbed_data/Fischer498/metadata_src/GSE62564_series_matrix.txt.gz",'rt',encoding="UTF-8"),
                                 skiprows=51,skipfooter=102-74,engine='python')
    metadatasurv.index = metadatasurv['!Sample_geo_accession'].apply(lambda x: x.replace('!',''))
    del metadatasurv['!Sample_geo_accession']
    metadatasurv = metadatasurv.T
    metadatasurv.columns = range(len(metadatasurv.columns))
    metadatasurv = metadatasurv[list(range(7,21))]
    metadatasurv.columns = [v.split(':')[0].replace(' ','_') for v in metadatasurv.ix[metadatasurv.first_valid_index()]]
    metadatasurv = metadatasurv.applymap(lambda x: x.split(': ')[1] if not x is np.nan else x)
    metadatasurv.index = metadata.index #Both sample sets are sorted the same way, but different gse names
    assert sum(metadatasurv.age == metadata.age_at_diagnosis) == len(metadata)
    metadata['overall_survival'] = metadatasurv.os_day.apply(int)
    metadata['eventfree_survival'] = metadatasurv.efs_day.apply(int)
    del metadatasurv
    metadata.Sample_geo_accession = metadata.Sample_geo_accession.apply(lambda x: x.lower())
    metadata.set_index("Sample_geo_accession",inplace=True)
    metadata.Sample_title = metadata.Sample_title.apply(lambda x: x[5:].replace(' patient ',''))
    metadata.death_from_disease = metadata.death_from_disease == '1'
    metadata.progression = metadata.progression == '1'

    #Expression data
    ## Array expression
    exprdata_A = pd.read_table(datadir+'R2_grabbed_data/Fischer498/GSE49710_R2.txt')
    exprdata_A.index = exprdata_A.pop('#H:hugo')
    del exprdata_A['probeset']
    
    ## RNAseq gene expression
    exprdata_G = pd.read_table(
        gzip.open(
            datadir+'R2_grabbed_data/Fischer498/sequencedData/GSE49711_SEQC_NB_TUC_G_log2.txt.gz',
            'rt',encoding="UTF-8"
        ),
        index_col = '00gene_id'
    )
    exprdata_G.columns = [
        metadata.reset_index().set_index('Sample_title').ix[c.split('_')[1]].Sample_geo_accession
        for c in exprdata_G.columns
    ]

    ## RNAseq transcript level expression
    exprdata_T = pd.read_table(
        gzip.open(
            datadir+'R2_grabbed_data/Fischer498/sequencedData/GSE49711_SEQC_NB_TUC_T_log2.txt.gz',
            'rt',encoding="UTF-8"
        ),
        index_col = '00transcript_id'
    )
    exprdata_T.columns = exprdata_G.columns

    ## RNAseq junction level expression
    exprdata_J = pd.read_table(
        gzip.open(
            datadir+'R2_grabbed_data/Fischer498/sequencedData/GSE49711_SEQC_NB_TUC_J_log2.txt.gz',
            'rt',encoding="UTF-8"
        ),
        index_col = 'sample_ID'
    )
    exprdata_J.columns = exprdata_G.columns

    # Default expression dataset -> exprdata_G
    exprdata = exprdata_G
    
    #aCGH
    aCGH = pd.read_table(datadir+'R2_grabbed_data/Fischer498/SEQC_aCGH/SEQC_aCGH_all_146.txt')
    geosearch = metadata[['Sample_title']].copy()
    geosearch.reset_index(inplace=True)
    geosearch.set_index('Sample_title',inplace=True)
    aCGH.Sample = aCGH.Sample.apply(lambda x: geosearch.ix[x].Sample_geo_accession)
    del geosearch
    aCGH['log2ratio'] = (aCGH.CN/2).apply(np.log2)
    #Convert coordinates to hg38
    lo = LSD.get_lift19to38()
    aCGH['Start38'] = aCGH.T.apply(lambda x: lo.convert_coordinate(x.Chromosome,x.Start)).apply(lambda x: x[0][1] if x else np.nan)
    aCGH['End38'] = aCGH.T.apply(lambda x: lo.convert_coordinate(x.Chromosome,x.End)).apply(lambda x: x[0][1] if x else np.nan)
    del lo, aCGH['Start'], aCGH['End']
    aCGH = aCGH.dropna().copy()
    #Assign genes to regions
    genannot = LSD.get_ensemblGeneannot()
    aCGH['genes'] = aCGH.T.apply(lambda x: {f.attributes['gene_name'][0] for f in genannot.region('{}:{}-{}'
                                                    .format(x.Chromosome[3:],int(x.Start38),int(x.End38)),featuretype='gene')})
    aCGH['nrGenes'] = aCGH.genes.apply(len)
    del genannot
    #To set cut offs look at hist => aCGH.log2ratio.hist(ax=ax,bins='auto')
    aCGH['annotation'] = aCGH.log2ratio.apply(lambda x: 'gain' if x > 0.3 else ('loss' if x < -0.3 else 'normal'))

    # Filter patients whom according to metadata survived, but had last follow up before 5 years
    metadata = metadata[metadata.death_from_disease |
        (~metadata.death_from_disease &
        (metadata.overall_survival > filteredOn['minimalSurvivalLastFollowup']))]
    exprdata_A = exprdata_A[metadata.index]
    exprdata_G = exprdata_G[metadata.index]
    exprdata_T = exprdata_T[metadata.index]
    exprdata_J = exprdata_J[metadata.index]
    aCGH = aCGH[aCGH.Sample.isin(metadata.index)]
    
    return Dataset(**locals())

def get_gtex():
    """
    Reference: https://www.gtexportal.org/home/datasets
    """
    gtex = pd.read_table(
        gzip.open(datadir+'gtexportal/GTEx_Analysis_v6p_RNA-seq_RNA-SeQCv1.1.8_gene_median_rpkm.gct.gz'),
        skiprows=2,index_col='Description'
    )
    ensgref = gtex.pop('Name')

    return Dataset(**locals())



# -*- coding: utf-8 -*-
"""bidali utilities

Functions for transforming pd.DataFrame and other generic data processing steps.id
"""
import pandas as pd

def unfoldDFlistColumn(df,listColName):
    """Unfold a pd.DataFrame

    A dataframe containing a column with lists as values will be unfolded to
    a dataframe with a new row for every value in the list.

    Args:
        df (pd.DataFrame): DataFrame to unfold.
        listColName (str): Name of the column that has list values.
    """
    unfoldedListCol = df.apply(
        lambda x: pd.Series(x[listColName]),axis=1
    ).stack().reset_index(level=1, drop=True)
    unfoldedListCol.name = listColName
    unfoldedDF = df.drop(listColName, axis=1).join(unfoldedListCol)
    return unfoldedDF



# -*- coding: utf-8 -*-
"""entrez

References:
- https://www.ncbi.nlm.nih.gov/Web/Search/entrezfs.html 
- https://www.ncbi.nlm.nih.gov/home/develop/api/
- ftp://ftp.ncbi.nih.gov/gene/
"""
from bidali import LSD
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, pandas as pd
from io import TextIOWrapper, StringIO
from urllib.parse import parse_qsl

@retrieveSources
def get_refseq():
    """
    Source: ftp://ftp.ncbi.nih.gov/refseq/H_sapiens/RefSeqGene/gene_RefSeqGene
    """
    entrez = pd.read_table(processedDataStorage+'gene_RefSeqGene', index_col='GeneID')
    return entrez

@retrieveSources
def get_gene_info():
    """
    Source: ftp://ftp.ncbi.nih.gov/gene/DATA/GENE_INFO/Mammalia/Homo_sapiens.gene_info.gz

    All species -> ftp://ftp.ncbi.nih.gov/gene/DATA/gene_info.gz
    """
    gene_info = pd.read_table(
        TextIOWrapper(gzip.open(processedDataStorage+'Homo_sapiens.gene_info.gz')),
        index_col='GeneID'
    )
    return gene_info



#!/usr/bin/env python
from bidali import LSD
import pandas as pd, numpy as np

@LSD.storeDatasetLocally
def get_UHRprofiles():
    """
    PDP meta cnv dataset from Ultra High Risk project. 
    Samples are all high risk (not ultra in se, that was project aim)

    Source: ~/Dropbiz/Basecamp/BRIP1/2017/2017_003_Data_analysis_BRIP1_paper/profiles_UHR/
    """
    datacn = LSD.expanduser("~/Dropbiz/Basecamp/BRIP1/2017/2017_003_Data_analysis_BRIP1_paper/profiles_UHR/")
    centromereshg38 = LSD.get_centromeres()
    lo = LSD.get_lift19to38()
    
    # All samples => TODO filter patients that only have whole chromosome gains
    samples = pd.read_table(datacn+'samples_UHR.txt')
    profiles = pd.read_table(datacn+'profiles_UHR.txt')
    profiles.chromosome = profiles.chromosome.apply(str).apply(lambda x: 'X' if x=='23' else x)
    profiles['size'] = profiles['max']-profiles['min']
    profiles['min38'] = profiles.apply(lambda x: lo.convert_coordinate('chr{}'.format(x['chromosome']),x['min']),axis=1).apply(lambda x: x[0][1] if x else np.nan)
    profiles['max38'] = profiles.apply(lambda x: lo.convert_coordinate('chr{}'.format(x['chromosome']),x['max']),axis=1).apply(lambda x: x[0][1] if x else np.nan)
    print('alterations lost per',profiles[profiles[['min38','max38']].isnull().any(axis=1)].groupby('chromosome').size())
    profiles = profiles[~profiles[['min38','max38']].isnull().any(axis=1)]
    del profiles['min']
    del profiles['max']

    #Study only non-normal gains
    profiles = profiles[profiles.annotation != 'normal']

    #Annotate chr arm
    profiles['chrarm'] = profiles.apply(lambda x: 'chr{}{}'.format(x['chromosome'],'p' if x['max38'] < centromereshg38.ix['chr'+x['chromosome']]['left_base'] else
                                                                   ('q' if x['min38'] > centromereshg38.ix['chr'+x['chromosome']]['right_base'] else 'p+q')), axis=1)
    #Annotate in profiles mycn status
    samples.index = samples.Name
    profiles['MYCN'] = profiles.profile_id.apply(lambda x: samples.ix[x].MYCN)
    #profsperMYCNstatus = profiles[profiles.chromosome != '2'].groupby(('MYCN','profile_id')).size()
    #profsperMYCNstatus[0].mean()
    #(profsperMYCNstatus[0].var())**.5

    return LSD.Dataset(profiles=profiles,metadata=samples)
    



# -*- coding: utf-8 -*-
"""Publication crawler and text mining module


Functions for creating dictionaries and starting up text mining explorations.
"""
import pandas as pd, numpy as np

# Utilities for making dictionaries
##CIDcounter class to create CID (concept id) according to Adil's requirement that it
##is the same as TID reference symbol
class CIDcounter:
    def __init__(self,startvalue=0):
        import itertools as it
        self.counter = it.count(startvalue)
        
    def next(self,increment):
        if increment:
            self.value = next(self.counter)
        else: next(self.counter) #this step ensures a CID == TID for a new reference concept
        return self.value

def makeCIDandTID(df,conceptCol='symbol',aliasCol='alias'):
    """Make text mining dictionary with TID and CID set

    According to Adil Salhi's text mining requirements.

    Args:
        df (pd.DataFrame): Dataframe to transform.
        conceptCol (str): Column name of main concept/symbol.
        aliasCol (str): Column name where all aliases for the concept
          are gathered in a list, including the concept itself as first list member.
    """
    from bidali.util import unfoldDFlistColumn
    df = unfoldDFlistColumn(df,aliasCol)
    df.reset_index(inplace=True,drop=True)
    df.index.name = 'TID' #term id
    c = CIDcounter()        
    df.insert(0, 'CID', (~df[conceptCol].duplicated()).apply(c.next)) #need to work with `not` duplicated column!
    return df

## get icd9 info
def get_icd9info(icd9code,type='disease',includeShortName=True):
    """Get info related to icd9code

    Info:
        https://clinicaltables.nlm.nih.gov/

    Reference:
        https://clinicaltables.nlm.nih.gov/apidoc/icd9cm_dx/v3/doc.html

    Args:
        icd9code (str): Code to query.
        type (option): `disease` or `procedure`.
        includeShortName: Also include the short name in output.
    """
    import requests, json
    if type == 'disease':
        url = 'https://clinicaltables.nlm.nih.gov/api/icd9cm_dx/v3/search?terms={term}'
    elif type == 'procedure':
        url = 'https://clinicaltables.nlm.nih.gov/api/icd9cm_sg/v3/search?terms={term}'
    else: raise Exception('Wrong type specified, should be disease or procedure, not',type)
    if includeShortName: url+='&ef=short_name'
    r = requests.get(url.format(term = icd9code))
    return json.loads(r.content)
    
# Dictionary functions
def get_biomart_gene_dictionary():
    from bidali.LSD.dealer.ensembl import get_biomart
    from bidali.genenames import fetchAliases
    bm = get_biomart(atts=[
        'ensembl_gene_id',
        'ensembl_peptide_id',
        'pdb',
        'entrezgene',
        'hgnc_symbol',
    ])
    HGNC_aliases = bm['HGNC symbol'].apply(lambda x: [(a,x) for a in fetchAliases(x,unknown_action='list')])
    HGNC_aliases = pd.concat([pd.Series(dict(a)) for a in HGNC_aliases])
    HGNC_aliases = pd.Series(HGNC_aliases.index.values, index=HGNC_aliases)
    HGNC_aliases.index.set_names('HGNC symbol', inplace = True)
    HGNC_aliases.name = 'HGNC alias'
    return bm.join(other = HGNC_aliases, on = 'HGNC symbol')

def get_gene_dictionary():
    """Gene symbol dictionary.
    Uses genenames.org data.
    """
    from bidali.LSD.dealer import genenames
    from bidali.util import unfoldDFlistColumn
    gn = genenames.get_genenames()
    gn['alias'] = gn.T.apply(
        lambda x: [x.symbol, x['name']] + #optionally add other names such as protein names here
        (
            x.alias_symbol.split('|') if x.alias_symbol is not np.nan else []
        )
    )
    print('Original columns',gn.columns)
    gn = gn[['hgnc_id','symbol','alias','uniprot_ids']].copy()
    print('Columns kept:',gn.columns)
    gn = makeCIDandTID(gn)
    return gn

def get_gene_family_dictionary(family_ids):
    """Gene family gene symbol dictionary.
    Uses genenames.org data.
    
    Args:
        family_ids (int or list of ints): Either a single family id, or a list of family ids.

    Example:
        >>> gfd = get_gene_family_dictionary(588) # HLA family
    """
    from bidali.LSD.dealer import genenames
    from bidali.util import unfoldDFlistColumn
    gf = genenames.get_genefamilies()

    # rename some columns
    gf.rename(
        {
            'Approved Symbol': 'symbol',
            'Approved Name': 'name',
            'HGNC ID': 'hgnc_id',
            'Previous Symbols': 'previous'
        }, axis=1, inplace=True
    )
    
    # filter families
    if isinstance(family_ids, int): family_ids = [family_ids]
    gf = gf[gf['Gene family ID'].isin(family_ids)].copy()

    # make alias
    gf['alias'] = gf.T.apply(
        lambda x: [x.symbol, x['name'],]
        +
        (
            x.previous.split(', ') if x.previous is not np.nan else []
        )
        +
        (
            x.Synonyms.split(', ') if x.Synonyms is not np.nan else []
        )
    )
    print('Original columns', gf.columns)
    gf = gf[['hgnc_id','symbol','alias']].copy()
    print('Columns kept:',gf.columns)
    gf = makeCIDandTID(gf)
    return gf



# -*- coding: utf-8 -*-
"""Data from genenames.org

Reference: https://www.genenames.org/
"""
from bidali import LSD
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, pandas as pd
from io import TextIOWrapper, StringIO

@retrieveSources
def get_genenames():
    """
    Source: genenames.tsv ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/tsv/non_alt_loci_set.txt
    """
    return pd.read_table(
        os.path.join(processedDataStorage, 'genenames.tsv'), low_memory = False
    )#, index_col='GeneID')

@retrieveSources
def get_genefamilies():
    """
    Source: genefamilies.tsv https://www.genenames.org/cgi-bin/genefamilies/download-all/tsv
    """
    return pd.read_table(
        os.path.join(processedDataStorage, 'genefamilies.tsv'), low_memory = False
    )



# -*- coding: utf-8 -*-
"""Ontologies from BioPortal

Reference: https://bioportal.bioontology.org/
"""
from bidali import LSD
from bidali.config import secrets
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, pandas as pd
from io import TextIOWrapper, StringIO
import urllib.request, urllib.error, urllib.parse
import json

REST_URL = "http://data.bioontology.org"

def list_ontologies(printout=True):
    import os
    from pprint import pprint
    
    API_KEY = secrets.getsecret('bioportal') #need to register -> then https://bioportal.bioontology.org/account API key
    
    def get_json(url):
        opener = urllib.request.build_opener()
        opener.addheaders = [('Authorization', 'apikey token=' + API_KEY)]
        return json.loads(opener.open(url).read())
    
    # Get the available resources
    resources = get_json(REST_URL + "/")
    
    # Get the ontologies from the `ontologies` link
    ontologies = get_json(resources["links"]["ontologies"])
    
    # Get the name and ontology id from the returned list
    ontology_output = []
    for ontology in ontologies:
        ontology_output.append(f"{ontology['name']}\n{ontology['@id']}\n")
    
    # Print the first ontology in the list
    #pprint(ontologies[0])
    
    # Print/return the names and ids
    if printout:
        print("\n\n")
        for ont in ontology_output:
            print(ont)
    else: return ontology_output

def get_labels(acronym):
    """
    Reference:
        https://github.com/ncbo/ncbo_rest_sample_code/blob/master/python/python3/get_labels.py

    Example:
        get_labels('BRO')
    """
    API_KEY = secrets.getsecret('bioportal') #need to register -> then https://bioportal.bioontology.org/account API key    
    
    def get_json(url):
        opener = urllib.request.build_opener()
        opener.addheaders = [('Authorization', 'apikey token=' + API_KEY)]
        return json.loads(opener.open(url).read())
    
    # Get all ontologies from the REST service and parse the JSON
    ontologies = get_json(REST_URL+"/ontologies")
    
    # Iterate looking for ontology with acronym BRO
    bro = None
    for ontology in ontologies:
        if ontology["acronym"] == acronym:
            bro = ontology
    
    labels = []
    
    # Using the hypermedia link called `classes`, get the first page
    page = get_json(bro["links"]["classes"])
    
    # Iterate over the available pages adding labels from all classes
    # When we hit the last page, the while loop will exit
    next_page = page
    while next_page:
        next_page = page["links"]["nextPage"]
        for bro_class in page["collection"]:
            labels.append(bro_class["prefLabel"])
        if next_page:
            page = get_json(next_page)
    
    # Output the labels
    for label in labels:
        print(label)
    



# -*- coding: utf-8 -*-
"""Drugging databases

Drug names, medication and procedure links.
"""
from bidali import LSD
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, zipfile, pandas as pd
from io import TextIOWrapper, StringIO
from urllib.parse import parse_qsl

class DrugBank:
    """Convenience class to work with
    https://www.drugbank.ca drugbank xml.
    
    Mainly methods to access specific data
    """
    def __init__(self,drugtree):
        self.tree = drugtree
        self.root = self.tree.getroot()
        self.namespace_symbol = 'ns0'
        self.tree_namespace = {self.namespace_symbol:'http://www.drugbank.ca'}

    def __getitem__(self,xpath):
        #prepending namespace to all path elements
        xpath = '/'.join(
            ':'.join((self.namespace_symbol,d)) for d in xpath.split('/')
        )
        return self.root.findall(xpath, self.tree_namespace)

    def get_tags(self):
        return {
            e.tag.replace(
                '{'+self.tree_namespace[self.namespace_symbol]+'}',
                ''
            )
            for e in self.tree.iter()
        }

    def get_drugnameset(self):
        return {
            d.text.strip() for d in self['drug/name']
        }
        
@retrieveSources
def get_drugbank():
    """
    Info: https://www.drugbank.ca/docs/drugbank.xsd

    Source: drugbank_all_full_database.xml.zip $locked$https://www.drugbank.ca/releases/5-1-1/downloads/all-full-database
    Source: https://www.drugbank.ca/docs/drugbank.xsd
    """
    import xml.etree.ElementTree as ET:
    with zipfile.ZipFile(os.path.join(processedDataStorage,'drugbank_all_full_database.xml.zip')) as z:
        with z.open('full database.xml') as f:
            tree = ET.parse(f)
    return DrugBank(tree)



# -*- coding: utf-8 -*-
"""fegnome:Fisher exact genome module

Functions and classes for FE testing in relation to genomic context
"""
#TODO https://www.math.hmc.edu/funfacts/ffiles/10006.3.shtml -> proportional volume
#TODO http://pythonhosted.org/gseapy/gseapy_example.html#prerank-example

import pandas as pd
import numpy as np
from collections import namedtuple, OrderedDict
import matplotlib.pyplot as plt
import matplotlib.patches as ptch
from scipy.stats import fisher_exact

def genesetOverlap(geneset_a,geneset_b,universe,alternative='greater'):
    """Calculate Fisher exact value for overlap of two genesets

    Args:
        geneset_a (set): Geneset A.
        geneset_b (set): Geneset B.
        universe (int): The size of the universe from which the genesets are taken.

    Returns:
        (oddsratio, p_value) result from scipy.stats.fisher_exact

    Todos:
        * output venn/contingency table, jaccard index
        * write test with random model als negative control
    """
    # transform genesets to sets if not provided as such
    if not isinstance(geneset_a,set): geneset_a = set(geneset_a)
    if not isinstance(geneset_b,set): geneset_b = set(geneset_b)
    # prepare contingency table
    universe_over = universe - len(geneset_a | geneset_b)
    geneseta_unique = len(geneset_a - geneset_b)
    genesetb_unique = len(geneset_b - geneset_a)
    common_genes = len(geneset_a & geneset_b)
    # calculate fe score
    return fisher_exact(
        [
            [ universe_over, geneseta_unique ],
            [ genesetb_unique, common_genes ]
        ],
        alternative = alternative
    )

def enrichometer(ranks,genesUp,genesDown=None,universe=None,fexact_H1='two-sided',
                 reservoirR1=1.4,reservoirR2=None,fillingLinewidths=False,ax=None,title=None,
                 feminpv=.05,invertx=False,textrotation=0,fontsize=12,**kwargs):
    """Enrichometer plot

    Args:
        ranks (pd.Series): Series with ranked statistic.
        genesUp (set): Up or differentially regulated genes.
        genesDown (set): If not provided, considered as unitary geneset.
        universe (int or list): Total number of genes in the universe
            or list with all genes in the universe .
        fexact_H1 (str): greater, less, or two-sided
        reservoirR1 (float): Radius1 of reservoir
        reservoirR2 (float): Radius2 of reservoir, default 0.8
        fillingLinewidths (bool): Needs to be reimplemented, is not proportional
        ax (ax): ax for plot.
        title (str): Plot title.
        feminpv (float): Plot where enrichment is the strongest;
            provide as float, if fisher enrichment at strongest is not
            significant in respect to value provided nothing is plotted
        invertx (bool): Invert x axis.
        textrotation (float): Text rotation.
        fontsize (float): Font size.

    Returns:
        figure[, leadingEdgeGenes if feminpv]

    Note:
        *kwargs* are passed to `ax.eventplot`
    """
    #enrichometer settings
    axiscale = 10
    eventoffset = 0
    if reservoirR1 and reservoirR2:
        eventlen = np.pi*reservoirR1*reservoirR2*len(ranks)/(axiscale*len(genesUp))
    else: eventlen = 1
    reservoirR2 = reservoirR2 if reservoirR2 else .8
    reservoirR1 = (reservoirR1 if reservoirR1 else
                   axiscale*eventlen*len(genesUp)/
                   (np.pi*reservoirR2*len(ranks)))
    padding = .1
    if not ax:
        fig,ax = plt.subplots(figsize=(8,2))
    else: fig = ax.get_figure()
    ax.axis('off')
    if fillingLinewidths:
        raise NotImplemented #todo use rectangle patches, does not work with linewidths
        axPixels = ax.transData.transform([(0,1),(1,0)])-ax.transData.transform((0,0))
        linewidths = (axPixels[1,0]/fig.dpi)*axiscale/len(ranks)
    settings = {
        'colors':'r',
        }
    settings.update(kwargs)

    #Normalize the ranks
    minR,maxR = (ranks.min(),ranks.max())
    normalized = ((ranks-minR)*axiscale/(maxR-minR)-axiscale-reservoirR1)
    
    ax.eventplot(normalized[normalized.index.isin(genesUp)],
                 lineoffsets=eventoffset,linelengths=eventlen,**settings)
    ax.set_xlim((-axiscale-reservoirR1-padding,reservoirR1+padding))
    ax.set_ylim((eventoffset-max(eventlen/2,reservoirR2)-padding,eventoffset+max(eventlen/2,reservoirR2)+padding))

    #Draw thermometer
    overlap = sum(ranks.index.isin(genesUp))
    filled = 1-overlap/len(genesUp)
    emptyReservoir = reservoirR1*2-(reservoirR1*2*filled)
    ax.add_patch(ptch.Ellipse((0,eventoffset),width=2*reservoirR1,height=2*reservoirR2,facecolor='r',edgecolor='none'))   
    ax.add_patch(ptch.Rectangle((-reservoirR1,eventoffset-reservoirR2),
                                emptyReservoir,2*reservoirR2,facecolor='w',edgecolor='none'))
    ax.add_patch(ptch.Ellipse((0,eventoffset),width=2*reservoirR1,height=2*reservoirR2,facecolor='none',edgecolor='k'))
    ax.add_patch(ptch.Rectangle((-axiscale-reservoirR1,eventoffset-eventlen/2),axiscale,eventlen,facecolor='w',edgecolor='k'))

    #Annotations
    ax.annotate('{:g}'.format(minR),(-axiscale-reservoirR1,(eventoffset-eventlen/2)-padding),
                ha='right' if invertx else 'left',va='top',rotation=textrotation,size=fontsize)
    ax.annotate('{:g}'.format(maxR),(-reservoirR1,(eventoffset-eventlen/2)-padding),
                ha='left' if invertx else 'right',va='top',rotation=textrotation,size=fontsize)

    if universe:
        if type(universe) is list: raise NotImplemented
        odds,pval = fisher_exact([[overlap,len(ranks)],[len(genesUp)-overlap,universe]],alternative=fexact_H1)
        ax.annotate('{}\n{:.3g}'.format(len(genesUp),pval),(0,eventoffset),
                    ha='center',va='center',rotation=textrotation,size=fontsize)

    if feminpv:
        fenrichscores = fenrichmentscore(ranks,genesUp)
        pvmin = fenrichscores.pvalue.min()
        if pvmin <= feminpv:
            leadingEdgeGene = fenrichscores[fenrichscores.pvalue==pvmin].first_valid_index()
            #ax.eventplot((normalized.loc[leadingEdgeGene],),lineoffsets=eventoffset,linelengths=eventlen,
            #             color='g')
            ax.add_patch(ptch.Rectangle((-axiscale-reservoirR1,eventoffset-eventlen/2),
                                        normalized.loc[leadingEdgeGene]+axiscale+reservoirR1,
                                        eventlen,facecolor='r',alpha=.4))
            ax.annotate('{:.3g}'.format(pvmin),(normalized.loc[leadingEdgeGene]+padding,eventoffset),
                        ha='left',va='center',size=fontsize)
        else: leadingEdgeGene = None
        
    if title:
        #ax.set_title(title)
        ax.annotate(title,(-axiscale-reservoirR1,eventoffset+padding+eventlen/2),
                    ha='right' if invertx else 'left',va='bottom',size=fontsize+2)

    if invertx: ax.invert_xaxis()
    
    return (fig,leadingEdgeGene) if feminpv else fig

def fenrichmentscore(ranks,genesUp,genesDown=None):
    """Calculate fe score for each position in the ranked list

    Args:
        ranks (pd.Series): Ranks for all genes.
        genesUp (set): Up or differentially regulated genes.
        genesDown (set): Optionally down regulated genes.

    Returns:
        pd.DataFrame with oddratio's and pvalue's.
    """
    genes = {g for g in genesUp if g in ranks.index}
    ranks = ranks.sort_values()
    fescores = []
    for i in range(1,len(ranks)):
        genesUpto,genesAfter = set(ranks.index[:i]),set(ranks.index[i:])
        fescores.append(
            fisher_exact([[len(genesUpto&genes),len(genesUpto-genes)],
                          [len(genesAfter&genes),len(genesAfter-genes)]],
                         alternative='greater')
        )
        
    return pd.DataFrame(fescores,columns=['oddratio','pvalue'],index=ranks.index[:-1])

RankSumResult = namedtuple('RankSumResult',[
    'probability',
    'genesetsize',
    'overlapsize',
    'rankedgenes',
    'nulldistrosize',
    'fe_s','fe_p'])
        
def rankSumProbability(ranks, geneset, alternative='greater',
                       nulldistrosize=10000, nulldistrosave=True, universe=30000, verbose=False):
    """Probability of the ranksum

    Args:
        ranks (pd.Series): Ranks of the genes.
        geneset (set): Genes for the rank sum.
        alternative (str): greater or less.
        nulldistrosize (int): Size of the null distribution to calculate.
        nulldistrosave (bool): Option to cache nulldistro for a given geneset size and ranks.
        universe (int): Size of the genes universe.
        verbose (bool): Extra info.

    Returns:
        namedtuple RankSumResult

    Example:
        >>> from .tests.test_fegnome import testRanks, testGeneset
        >>> rankSumProbability(testRanks,testGeneset,nulldistrosave=False) # doctest: +ELLIPSIS
        RankSumResult(...)
    """
    realsetlen = len(geneset)
    setlen = ranks.index.isin(geneset).sum()
    if not setlen: return RankSumResult(None,realsetlen,setlen,None,None,None,None)
    ranksum = ranks[ranks.index.isin(geneset)].sum()
    if nulldistrosave:
        import pickle,hashlib
        from bidali import LSD
        hashindex = hashlib.md5(str(tuple(sorted(ranks))).encode()).hexdigest()
        try: nulldistros = pickle.load(open(LSD.processedDataStorage+'ranklogsum_nulldistros.pickle','rb'))
        except FileNotFoundError: nulldistros = {}
        if (hashindex,setlen) in nulldistros and nulldistros[(hashindex,setlen)][0] >= nulldistrosize:
            nulldistrosize,nulldistro = nulldistros[(hashindex,setlen)]
            if verbose: print('Using previously stored nulldistribution size',nulldistrosize)
        else:
            nulldistro = pd.Series([ranks.sample(n=setlen).sum() for i in range(nulldistrosize)])
            nulldistros[(hashindex,setlen)] = (nulldistrosize, nulldistro)
            pickle.dump(nulldistros,open(LSD.processedDataStorage+'ranklogsum_nulldistros.pickle','wb'))
    else:
        nulldistro = pd.Series([ranks.sample(n=setlen).sum() for i in range(nulldistrosize)])

    if alternative == 'greater':
        probability = float(sum(nulldistro < ranksum))/nulldistrosize
    elif alternative == 'less':
        probability = float(sum(nulldistro > ranksum))/nulldistrosize

    fe = fisher_exact([[setlen,len(ranks)],
                       [realsetlen-setlen,universe]],
                      alternative='greater')
        
    return RankSumResult(probability,realsetlen,setlen,len(ranks),nulldistrosize,*fe)



# -*- coding: utf-8 -*-
"""Wellcome Sanger Institute datasets

Reference: http://www.sanger.ac.uk/
"""
from bidali import LSD
import os, gzip, pandas as pd

@LSD.retrieveSources
def get_census():
    """Cancer census genes

    Info: https://cancer.sanger.ac.uk/census
    Locked source: https://cancer.sanger.ac.uk/cosmic/s3download?data=GRCh38%2Fcosmic%2Fv84%2Fcancer_gene_census.csv
    """
    return pd.read_csv(os.path.join(LSD.processedDataStorage,'cancer_gene_census.csv'),index_col=0)



#!/usr/bin/env python
from bidali import LSD
import pandas as pd, numpy as np

@LSD.storeDatasetLocally
def get_THMYCN():
    """
    Source: ~/LSData/private/2015_Hyperplasia_ABC/uniData.tx
    Source: ~/LSData/private/2015_Hyperplasia_ABC/SampleAnnotation.txt

    TODO refactor using retro DEA function
    """
    from bidali.retro import importr, ro, pandas2ri, base
    #TH-MYCN mouse model incorporation
    limma = importr('limma')
    prepcore = importr('preprocessCore')
    thannot = pd.read_table(LSD.datadir+'2015_Hyperplasia_ABC/SampleAnnotation.txt')
    thannot.group = thannot.group.apply(lambda x: x.replace('.','_'))
    thdata = pd.read_table(LSD.datadir+'2015_Hyperplasia_ABC/uniData.txt',sep=' ')
    thdata.columns = thannot.group + (['_r'+str(i+1) for i in range(4)]*3*2)
    genotype = thannot.genotype
    age = thannot.age
    genotype_r = ro.r.relevel(ro.FactorVector(genotype),ref='WT')
    thdatanorm_r = prepcore.normalize_quantiles(base.as_matrix(thdata))
    thdatanorm_r.colnames = ro.StrVector(thdata.columns)
    thdatanorm_r.rownames = ro.StrVector(thdata.index)
    metadata_r = base.data_frame(genotype=genotype_r,age=age)
    designSamples_r = ro.r['model.matrix'](ro.r.formula('~genotype:age'),data=metadata_r)
    fit_r = limma.lmFit(thdatanorm_r,designSamples_r)
    fit_r = limma.eBayes(fit_r)
    coefficients_r = fit_r.rx2('coefficients')
    thcoeffs = pd.DataFrame(
        pandas2ri.ri2py(coefficients_r),
        index=coefficients_r.rownames,
        columns=coefficients_r.colnames
    )
    thcoeffs['lineardiff'] = thcoeffs['genotypeTG:age']-thcoeffs['genotypeWT:age']

    return LSD.Dataset(exprData=thdata,
                       metadata=thannot,
                       lmCoeffs=thcoeffs)



#!/usr/bin/env python
from bidali import LSD
import pandas as pd, numpy as np, gzip
from os.path import expanduser, exists
from itertools import count
from bidali.LSD import storeDatasetLocally, datadir, Dataset

def get_NB39():
    """
    39 neuroblastoma cell lines + RPE1 and HU.FETAL.BRAIN

    Reference: https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE89413
    Source: https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE89413&format=file&\
file=GSE89413%5F2016%2D10%2D30%2DNBL%2Dcell%2Dline%2DSTAR%2Dfpkm%2Etxt%2Egz
    """
    exprdata = pd.read_table(
        gzip.open(
            datadir+'GEO/NB39_celllines_Maris/GSE89413_2016-10-30-NBL-cell-line-STAR-fpkm.txt.gz',
            'rt',encoding='UTF-8'
        ),
        index_col='GeneID'
    )

    return Dataset(exprdata=exprdata)

def get_CCLE():
    """
    Reference: https://portals.broadinstitute.org/ccle
    """
    metadata = pd.read_table(datadir+'CCLE/CCLE_sample_info_file_2012-10-18.txt',index_col='CCLE name')
    exprdata = pd.read_table(datadir+'CCLE/CCLE_Expression_Entrez_2012-09-29.gct',skiprows=2,index_col='Description')
    return Dataset(exprdata=exprdata,metadata=metadata)



#!/usr/bin/env python
from bidali import LSD
import gzip, pandas as pd
from io import TextIOWrapper

def get_cancerrxgene(datadir=LSD.datadir+'PMID_published_tables/27397505_pharmacogenomics_cancer/', gene_cn=False, exprdata=False):
    """
    Reference: http://www.cancerrxgene.org/downloads

    cn data:
    Four comma seperated peices of data are presented for each gene cell line combination (n1,n2,n3,n4), hyphen (-) is used where the value is unknown:-

    n1 - Maximum copy number of any genomic segment containing coding sequence of the gene (-1 indicates a value could not be assigned).
    n2 - Minimum copy number of any genomic segment containing coding sequence of the gene (-1 indicates a value could not be assigned).
    n3 - Zygosity - (H) if all segments containing gene sequence are heterozygous, (L) if any segment containing coding sequence has LOH, 
         (0) if the complete coding sequence of the gene falls within a homozygous deletion.
    n4 - Disruption (D) if the gene spans more than 1 genomic segment (-) if no disruption occures.

    >>> cdx = get_cancerrxgene()
    """
    # Metadata
    cell_lines = pd.read_excel(datadir+'Cell_Lines_Details.xlsx',sheetname='Cell line details')
    cell_tissues = pd.read_excel(datadir+'Cell_Lines_Details.xlsx',
                                 sheetname='COSMIC tissue classification',index_col='Line')
    cell_lines = cell_lines.join(cell_tissues,on='Sample Name')
    del cell_lines['COSMIC_ID']
    del cell_tissues

    # Compounds
    compounds = pd.read_excel(datadir+'Screened_Compounds.xlsx',index_col='Drug Name')

    # Compound responses
    compound_responses = pd.read_excel(datadir+'v17_fitted_dose_response.xlsx',index_col='COSMIC_ID')

    # Expression data
    if exprdata:
        exprdata = pd.read_table(TextIOWrapper(gzip.open(datadir+'sanger1018_brainarray_ensemblgene_rma.txt.gz')),index_col='ensembl_gene')
    
    # CN data
    if gene_cn:
        gene_cn = pd.read_excel(datadir+'Gene_level_CN.xlsx',sheetname='Gene_level_CN',index_col='gene')

    return LSD.Dataset(**locals())



# -*- coding: utf-8 -*-
"""Bidali configuration

This module organizes the configuration for the general bidali package and
the LSD subpackage.

Todo:
    * Give example symlinking privatedir ln -s ~/Dropbiz/Lab/z_archive/Datasets ~/LSData/private
"""

import configparser, os
configFileOptions = [
    'bidali.cfg', # in current working dir
    os.path.expanduser('~/.bidali.cfg'),
    '/usr/local/etc/bidali.cfg'
]

# Default configuration
config = configparser.ConfigParser()
config['bidali'] = {
    'plotting_backend': 'TkAgg'
}

config['LSD'] = {
    'cachetime': '4w', #supports w[eeks], d[ays] or h[ours]
    'cachedir': os.path.expanduser('~/LSData/cache/'),
    'privatedir': os.path.expanduser('~/LSData/private/')
}

config['plotting'] = {
    'mpl_backend': 'TkAgg',
    'interactive': 'yes',
    'use_seaborn_bug_workaround': 'no'
}

# Read configuration file
for configFile in configFileOptions:
    if os.path.exists(configFile):
        config.read(configFile)
        break #only reads the first config file found

# Secrets: config for storing user API keys and other sensitive/personal information
from kindi import Secrets
secrets = Secrets(default_section=__package__)



# -*- coding: utf-8 -*-
"""Expression analysis module

Defines the Expan object for all your expression analysis needs.
Relies under the hood on the retro module and R packages.

Example:
    >>> from bidali.expan import Expan
    >>> expan = Expan(                                                                                                      
    ...   counts = '/home/christophe/Dropbiz/Basecamp/BRIP1/2016/2016_008_DDR_RNA sequencing/raw data-results/2015_TMPYP4_SVH_RSEMcounts_pp.csv',
    ...   metadata = '/home/christophe/Dropbiz/Basecamp/BRIP1/2016/2016_008_DDR_RNA sequencing/raw data-results/2015_TMPYP4_metadata.csv'
    ... )
    >>> expan.designator(design = '~batch+treatment+cellline', reflevels = {'batch':'seq1','cellline':'IMR32','treatment':'control'})
    >>> expan.exdif(contrasts = [4, 5], countfilter = 1)
    >>> brip1 = expan['BRIP1']
    >>> brip1.plotCounts('cellline','treatment')
"""
import pandas as pd
import numpy as np
from bidali import retro
from collections import OrderedDict

class Expan:
    """Expression analysis class

    This class is the starting point for a bidali expression analysis workflow.
    """

    def __init__(self,
                 counts, metadata, export = None,
                 annotations = None, annotatedOnly = True,
                 counts_kwargs={}, metadata_kwargs={}, **kwargs):
        """Expression analysis class

        Args:
            counts (str): The filename for the counts table.
            metadata (str): The filename for the metadata table.
            export (str): Export location. Not implemented yet!
            annotations (pd.DataFrame): Annotations for the counts
                gene identifiers. Default loads biomart table.
            annotatedOnly (bool): Only analyze genes that have annotation.
            counts_kwargs (dict):
            metadata_kwargs (dict):
                Common counts and metadata pd.read_table arguments can 
                be provided as extra key word arguments. Different arguments
                for pd.read_table need to be specified in the counts_kwargs
                and metadata_kwargs dictionaries. The default index col is 0
                and the default sep is ','. 
        """
        # Include common kwargs in counts and metadata specific kwargs
        default_kwargs = {'index_col':0,'sep':','}
        default_kwargs.update(kwargs)
        counts_kwargs.update(default_kwargs)
        metadata_kwargs.update(default_kwargs)
        # Prepare tables
        self.counts = pd.read_table(counts,**counts_kwargs)
        ## Prepare metadata
        if isinstance(metadata,str):
            self.metadata = pd.read_table(metadata,**metadata_kwargs)
        else: # prepare metadata from counts columns
            if 'rename_cols' in metadata:
                rename_cols = metadata.pop('rename_cols')
                self.counts.columns = [rename_cols(c) for c in self.counts.columns]
            self.metadata = pd.DataFrame({
                m:[metadata[m](c) for c in self.counts.columns] for m in metadata
            }, index = self.counts.columns)
        if len(self.metadata) != sum(self.counts.columns.isin(self.metadata.index)):
            raise Exception('count column names and metatada row names need to match!')
        # Set column order of counts, to row order of metadata
        self.counts = self.counts[self.metadata.index]
        self._robjects = {} #dict to save private R objects
        if annotations is not None: self.annotations = annotations
        else:
            from bidali.LSD.dealer.ensembl import get_biomart
            biomart = get_biomart()
            biomart = biomart[~biomart['Gene stable ID'].duplicated()]
            biomart = biomart.set_index('Gene stable ID')
            self.annotations = biomart
        if annotatedOnly:
            self.counts = self.counts[self.counts.index.isin(self.annotations.index)]
        if export:
            self.exporter(export)

    def __getitem__(self,key):
        """
        First assumes a gene name, and queries the different tables
        for all gene-related information
        """
        if key in self.annotations.index:
            geneid = key
        else:
            geneid = self.annotations.index[self.annotations['Gene name'] == key]
        return GeneResult(
            name = key,
            counts = self.counts.loc[geneid],
            counts_norm = self.counts_norm.loc[geneid],
            exdif = OrderedDict([(k,self.results[k].loc[geneid]) for k in self.results]),
            parent = self
        )

    def __delitem__(self,sample):
        """
        Deletes a sample from counts and metadata
        """
        self.counts = self.counts.drop(sample,axis=1)
        self.metadata = self.metadata.drop(sample)
        

    def designator(self, design, reflevels):
        """Prepare experimental design

        Args:
            design (str): The design of the experiment, e.g. '~treatment+batch'
            reflevels (dict): The reference level for each factor of the metadata

        Example:
            reflevels = {'treatment':'SHC002_nox','rep':'rep1'}
        """
        self.design = design
        self._robjects['design'], self.designmatrix = retro.prepareDesign(
            self.metadata[self.metadata.columns[self.metadata.columns.isin(reflevels)]],
            self.design, reflevels, RReturnOnly = False
        )

    def exdif(self, contrasts, countfilter=1, quantro = False):
        """Differential expression method

        Runs differential expression workflow

        Args:
            contrasts (list or dict): If design coefficients suffice provide
                a list of the design column numbers of interest, else provide
                a dict with contrast name and contrast constructor string.
            countfilter (int): Mininum average number of reads per sample.
                Genes that do not meet this requirement are filtered.
            quantro (bool): Include quantro analysis. Not implemented yet!

        Examples:
            >>> contrasts = [1,2]
            >>> contrasts = {
            ...   'c1': 'treatmentSHC002_dox - treatmentTBX2sh25_dox','treatmentSHC002_dox - treatmentTBX2sh27_dox',
            ...   'c2': 'treatmentSHC002_nox - treatmentTBX2sh25_nox','treatmentSHC002_nox - treatmentTBX2sh27_nox'
            ...  }
        """
        self.contrasts = contrasts
        self.counts_fltd = (
            self.counts[self.counts.T.sum() > countfilter * len(self.counts.columns)]
            if countfilter else self.counts
        )
        self.results, self.counts_norm = retro.DEA(self.counts_fltd, self._robjects['design'], self.contrasts)
        for k in self.results:
             self.results[k]['Gene name'] = self.annotations['Gene name']

    def gsea(self, genesets = None, exclude_results = {}):
        """Geneset enrichment analysis

        Expan.exdif should have already been executed. The gene exdif results will be deduplicated.
        As such, genes that have multiple locations in the genome will only maintain there best result
        in the geneset enrichment analysis. This creates a small bias for such genes.

        Args:
            genesets (dict): Should be a dict with string keys designating the geneset names,
                and gene list values containing all the geneset members. If not provided all
                MSigDB geneset collections are used.
            exclude_results (set): A set of exdif results for which the gsea analysis is not required.
        Returns:
            gsea result
        """
        from bidali.retro import genesets2indices_r, fgsea
        if genesets:
            gscollections = {'custom': genesets}
        else:
            from bidali import LSD
            gscollections = LSD.get_msigdb6()
        # deduplicate gene exdif results
        results_deduplicated = {
            r:self.results[r][~self.results[r]['Gene name'].duplicated()]
            for r in self.results if r not in exclude_results
        }
        self.gsea_results = {}
        for r in results_deduplicated:
            geneLabels = results_deduplicated[r]['Gene name'].as_matrix()
            gscollections_r = {
                gsc:genesets2indices_r(gscollections[gsc], geneLabels)
                for gsc in gscollections
            }
            self.gsea_results[r] = {
                gsc:fgsea(
                    gscollections_r[gsc],
                    results_deduplicated[r].set_index('Gene name').t,
                    minSize = 15
                )
                for gsc in gscollections   
            }

    def exporter(self, location):
        """Export expression analysis object
        
        All relevant tables and figures are saved in a zipfolder at the specified location.
        """
        pass

    @staticmethod
    def importer(location):
        """Import tables from a previous Expan

        Only counts, metadata and annotation are imported.
        """
        pass

    @staticmethod
    def convert_excell_to_csv(filename,sheet_name=0):
        """
        Convenience function to e.g. convert an metadata excell table to a csv equivalent

        Args:
            filename (str): Excell filename.
            sheet_name (int or str): Either sheet number or sheet name.

        Returns:
            filename of converted table
        """
        newFilename = filename[:filename+rindex('.')]+'.csv'
        pd.read_excel(filename,sheet_name=sheet_name).to_csv(newFilename)
        return newFilename

class GeneResult:
    """
    Collects gene specific results.
    Methods for plotting, calculating knockdown
    """
    def __init__(self, name, counts, counts_norm, exdif, parent):
        """
        Differential expression gene result
        """
        self.name = name
        self.counts = counts
        self.counts_norm = counts_norm
        self.exdif = exdif
        self.parent = parent

    def __repr__(self):
        return '<GeneResult {}>'.format(self.name)

    def __str__(self):
        return '{}:\nCounts:\n{}\n\nNormalised:\n{}\n\nResults:\n{}\n'.format(
            self.name,
            self.counts.T,
            self.counts_norm.T,
            '\n'.join(['{} =>\n{}\n'.format(k,self.exdif[k].T) for k in self.exdif])
        )

    def calcKD(self, groupingCol, knockdownCol, knockdownControl = 'control'):
        """Calculate knockdown

        Args:
            groupingCol (str): Grouping column name in analysis metadata.
            knockdownCol (str): KD column name in analysis metadata.
            knockdownControl (str): Within knockdownCol the value of the control condition

        Returns:
            float

        Note:
            * Missing values should be provided as np.NaN
        """
        treatments = {t for t in self.parent.metadata[knockdownCol] if t != knockdownControl}
        kd = self.counts.groupby(self.parent.metadata[groupingCol]
        ).aggregate(
            {t:lambda x,t=t: (x.filter(regex=t).get_values()[0] if x.filter(regex=t).any() else np.nan)/
             x.filter(regex=knockdownControl).get_values()[0]
             for t in treatments})
        return kd

    def plotCounts(self, x, hue, normalisedCounts = True, **kwargs):
        """
        Plot counts for a gene
        """
        from .visualizations import plotGeneCounts
        df = (self.counts_norm if normalisedCounts else self.counts).T.join(self.parent.metadata)
        df.columns = ['counts'] + list(df.columns[1:])
        return plotGeneCounts(df, x=x, hue=hue, **kwargs)

class GenesetResult(GeneResult):
    """
    Inherits from GeneResult.
    Offers extra methods for geneset related functionality
    """
    def __repr__(self):
        return '<GenesetResult {}>'.format(self.name)



from setuptools import setup, find_packages

setup(name = 'bidali',
      version = '0.1.16',
      description = 'Biological Data Analysis Library',
      url = 'https://github.com/dicaso/bidali',
      author = 'Christophe Van Neste',
      author_email = 'christophe.vanneste@ugent.be',
      license = 'MIT',
      packages = find_packages(),
      python_requires='>3.6',
      install_requires = [
          #Generated with `pipreqs .` and then moved here from requirements.txt
          'kindi',
          'networkx',
          'biomart',
          'pandas',
          'xarray',
          'gffutils',
          'scipy',
          'numpy',
          'pyliftover',
          'requests',
          'seaborn',
          'setuptools',
          'matplotlib',
          'plumbum',
          'lifelines',
          'tzlocal'
      ],
      extras_require = {
          'retro': ['rpy2'],
          #for netCDF4 output, depends on: libnetcdf-dev libhdf5-dev
          'netcdf': ['netCDF4'],
          'documentation': ['Sphinx']
      },
      zip_safe = False,
      #entry_points = {
      #    'console_scripts': ['getLSDataset=LSD.command_line:main'],
      #},
      test_suite = 'nose.collector',
      tests_require = ['nose']
)

#To install with symlink, so that changes are immediately available:
#pip install -e . 



# -*- coding: utf-8 -*-
"""ensembl

Reference: https://www.ensembl.org/
"""
from bidali import LSD
from bidali.LSD import retrieveSources,cacheableTable,processedDataStorage,datadir
import os, gzip, pandas as pd
from io import TextIOWrapper, StringIO
from urllib.parse import parse_qsl

## Biomart
@cacheableTable
def get_biomart(atts=None,dataset='hsapiens_gene_ensembl'):
    """
    Get biomart id mappings

    Info:
        For possible attributes, see attribute table in
        https://www.bioconductor.org/packages/2.7/bioc/vignettes/biomaRt/inst/doc/biomaRt.pdf

    >>> bmrt = get_biomart()
    """
    import biomart
    server = biomart.BiomartServer("http://www.ensembl.org/biomart")
    # default atts
    if not atts:
        atts = ['external_gene_name','external_gene_source','ensembl_gene_id',
                'ensembl_transcript_id','ensembl_peptide_id']
    seda = server.datasets[dataset]
    s = seda.search({'attributes': atts}, header=1)
    data = pd.read_table(StringIO(s.content.decode()))
    return data

## Human resources
def get_ensembl(onlyGeneLabeled=True,onlyInChromosomes=None):
    import warnings
    warnings.warn("deprecated, use get_biomart instead", DeprecationWarning)
    ensembl = pd.read_table(datadir+'Genomes/Ensembl/Biomart/idmapping_extended.txt',
                        header=None,index_col=0,names=('egid','etid','start','stop','gcC','chr','strand','TSS','typeg','typet','gene_label','entrez'))
    ensembl = ensembl[~ensembl.index.duplicated()]
    if onlyGeneLabeled: ensembl = ensembl[ensembl.gene_label.isnull().apply(lambda x: not(x))]
    ensembl.chr = ensembl.chr.apply(lambda x: 'chr'+x.replace('MT','M'))
    if onlyInChromosomes: ensembl = ensembl[ensembl.chr.isin(onlyInChromosomes)]
    return ensembl

@retrieveSources
def get_ensemblGeneannot():
    """
    Info: http://www.ensembl.org/info/data/ftp/index.html
    Source: ftp://ftp.ensembl.org/pub/release-88/gtf/homo_sapiens/Homo_sapiens.GRCh38.88.gtf.gz
    """
    import gffutils
    try: db = gffutils.FeatureDB(processedDataStorage+'Homo_sapiens.GRCh38.88.sqlite3')
    except ValueError:
        if not os.path.exists(processedDataStorage+'Homo_sapiens.GRCh38.88.gtf.gz'):
            raise FileNotFoundError
        db = gffutils.create_db(processedDataStorage+'Homo_sapiens.GRCh38.88.gtf.gz',
                                processedDataStorage+'Homo_sapiens.GRCh38.88.sqlite3',
                                disable_infer_genes=True,disable_infer_transcripts=True)
    return db

## Cross species resources
### Mouse
@retrieveSources
def get_mouseEnsemblSet():
    """
    Source: ftp://ftp.ensembl.org/pub/release-90/gff3/mus_musculus/Mus_musculus.GRCm38.90.gff3.gz
    """
    db = pd.read_table(TextIOWrapper(gzip.open(processedDataStorage+'Mus_musculus.GRCm38.90.gff3.gz')),comment='#',
                       names='seqid,source,type,start,end,score,strand,phase,attribute'.split(','),low_memory=False)
    genes = db[db.type == 'gene'].copy()
    transcripts = db[db.type == 'mRNA'].copy()
    del db
    transcripts['TRANSCRIPT_ID'] = transcripts.attribute.apply(lambda x: parse_qsl(x)[0][1].split(':')[1])
    transcripts['GENE_ID'] = transcripts.attribute.apply(lambda x: parse_qsl(x)[1][1].split(':')[1])
    transcripts.set_index('TRANSCRIPT_ID',inplace=True)
    return (genes,transcripts)



# -*- coding: utf-8 -*-
"""LSD: Localised Structured Datasets
Module that preprocesses commonly used datasets
and makes them available for use in python.
"""

import gzip, pickle, time, os
from zipfile import ZipFile
from io import TextIOWrapper, StringIO
import pandas as pd, numpy as np, xarray as xr
from os.path import expanduser, exists
from collections import OrderedDict
from contextlib import redirect_stdout, redirect_stderr

## Defaults
from ..config import config
processedDataStorage = config['LSD']['cachedir']
datadir = config['LSD']['privatedir']

# Check file locations
if not (
        os.path.exists(datadir) and
        os.path.exists(processedDataStorage)
        ):
    import warnings
    warnings.warn(
        'Create directory %s or %s for bidali.LSD to function properly, or configure cachedir/privatedir'
        % (datadir, processedDataStorage)
    )

## Utility functions
def getLSDataset(name,**kwargs):
    try: return name(**kwargs)
    except TypeError:
        return globals()[name](**kwargs)

def listLSDatasets(subpackages=True):
    print(*[i for i in sorted(globals()) if i.startswith('get_')],sep='\n')
    if subpackages:
        print('\nWithin subpackages:')
        import pkgutil as pk
        from inspect import getmembers
        for mod in pk.walk_packages(__path__,__name__+'.'):
            if not mod.ispkg:
                mol = pk.importlib.import_module(mod[1])
                print('>',mod[1])
                print(*[i for i in sorted(dict(getmembers(mol)))
                        if i.startswith('get_')],sep='\n',end='\n\n')

# Structured dataset classes
## General
class Dataset(object):
    """A Dataset object is a collection of data tables, accessible as
    attributes from the Dataset object.

    to_R pushes the sub datasets to R making them available in the
    global namespace
    """
    def __init__(self,**kwargs):
        self.__datasets__ = set(kwargs)
        for kw in kwargs:
            self.__setattr__(kw,kwargs[kw])

    def generateRData(self):
        raise NotImplementedError

class IntegratedDataset(object):
    """IntegratedDataset checks and assures that for a set of samples, 
    features and datatypes all data is internally cross-referencable and
    then stores it as an xarray.
    """
    def __init__(self, xarray_kwargs = {}, **kwargs):
        self.data = xr.Dataset(data_vars = kwargs, **xarray_kwargs)

    def save(self,filename,csvsep=',',csvdec='.'):
        """Save IntegratedDataset to disk

        Iterates over the individual data variables and saves them
        in a zipfolder

        Args:
            filename (str): Save location. Extension '.zip' will be added.
        """
        import zipfile, io
        if filename.endswith('.zip'): filename = filename[:-4]
        with zipfile.ZipFile('{}.zip'.format(filename), mode='w') as zdir:
            for datatablename in self.data.data_vars:
                b = io.StringIO()
                self.data.data_vars[datatablename].to_pandas().to_csv(b,sep=csvsep,decimal=csvdec)
                b.seek(0)
                zdir.writestr('{}.csv'.format(datatablename),b.read())
                
    @classmethod
    def load(cls,filename,csvsep=',',csvdec='.'):
        """Load IntegratedDataset from disk zip folder.

        Args:
            filename (str): Save location. Extension '.zip' will be added.
        """
        import zipfile, io
        if filename.endswith('.zip'): filename = filename[:-4]
        with zipfile.ZipFile('{}.zip'.format(filename), mode='r') as zdir:
            tables = {
                zf.filename[:-4]:pd.read_csv(
                    io.TextIOWrapper(zdir.open(zf)),
                    sep=csvsep, decimal=csvdec, index_col=0
                )
                for zf in zdir.infolist()
            }
        return cls(**tables)

    # def save_netcdf(self, filename):
    #     """Save IntegratedDataset to disk

    #     Uses xr.Dataset netCDF method

    #     Args:
    #         filename (str): Save location. Suggested extension '.nc'.
    #     """
    #     raise NotImplementedError('Issues with netcdf output')
    #     self.data.to_netcdf(filename)

    # @staticmethod
    # def load_from_netcdf(filename):
    #     """Load IntegratedDataset from disk

    #     Uses xr.Dataset netCDF method

    #     Args:
    #         filename (str): Filename of dataset to load.
    #     """
    #     raise NotImplementedError('serializing and loading not yet working')
    #     return xr.open_dataset(filename)
    

## Specific integrated datasets
class CoexDataset(IntegratedDataset):
    """IntegratedDataset that contains data on
    copy number alteration status of genes (optionally
    including the regions), and gene expression data.

    *conureal* is optional, but if provided and *conugeal* is None,
    *conugeal* is calculated based on *conureal* and the genes in
    *expression*.

    Args:
        conugeal (pd.DataFrame): Copy number gene alterations, samples should be columns and genes rows.
        expression (pd.DataFrame): Expression data, samples should be columns and genes rows.
        metadata (pd.DataFrame): Annotation data should be columns and samples rows.
            Recommended columns are `DOD` (bool, "dead of disease"), `POD` (bool, "progress of disease"), `overallSurvival` (int), `progressionFreeSurvival` (int)
        metadataColMapping (dict): dict if metadata column names need to be changed to follow recommendations, e.g. {'survived':'DOD'}.
        conureal (pd.DataFrame): Copy number region alterations. 
            Required columns are `chrom`, `chromStart`, `chromEnd` and `samples`.
        conurealColMapping (dict): dict if conureal column names need to be changed to follow requirements , e.g. {'chr':'chrom'}
        filterUnknownExpressed (bool): If True, genes for which there is no expression data, will be filtered from conugeal.
    """
    def __init__(self,conugeal,expression,metadata,metadataColMapping={},conureal=None,conurealColMapping={},filterUnknownExpressed=True):
        # Applying col mapping
        if metadataColMapping: metadata = metadata.rename(metadataColMapping,axis=1)
        if conurealColMapping: conureal = conureal.rename(conurealColMapping,axis=1)

        # Checking data
        import warnings
        from pandas.api import types
        typefns = {bool:types.is_bool_dtype,int:types.is_integer_dtype}
        for n,t in (
                ('DOD',bool),
                ('POD',bool),
                ('overallSurvival',int),
                ('progressionFreeSurvival', int)
        ):
            if n in metadata and not typefns[t](metadata[n]):
                warnings.warn('%s in metadata is not of correct type %s' % (n,t))
            elif n not in metadata:
                warnings.warn('%s not in metadata' % n)
        
        # If only regions is provided, and conugeal is None, calculate gene copy number values for all expression genes
        if conureal and not conugeal:
            conugeal = self.__class__.calculate_conugeal(conureal,expression.index)
        self.conureal = conureal

        # Filter genes for which there is no expression data
        if filterUnknownExpressed:
            conugeal = conugeal.filter(items=expression.index,axis=0)
            
        #Setting dimension names
        conugeal.columns.name = expression.columns.name = metadata.index.name = 'sample'
        conugeal.index.name = expression.index.name = 'gene'
        metadata.columns.name = 'annotation'
        if conureal:
            conureal.columns.name = 'sample'
            conureal.index.name = 'region'
            
        super().__init__(
            conugeal=conugeal,
            expression=expression,
            metadata=metadata
        )

    @staticmethod
    def calculate_conugeal(conureal,genes):
        # Assign genes to regions
        genannot = get_ensemblGeneannot()
        segments['genes'] = segments.T.apply(
            lambda x: {f.attributes['gene_name'][0] for f in genannot.region('{}:{}-{}'.format(
                x.chromosome[3:],int(x.Start38),int(x.End38)),featuretype='gene')}
        )
        segments['nrGenes'] = segments.genes.apply(len)
        del genannot#, get_genes
        conugeal = {}
        for sname,sample in segments.groupby('sample'):
            conugeal[sname] = {}
            for i in sample.T:
                for g in sample.ix[i].genes: conugeal[sname][g] = sample.ix[i].value
        conugeal = pd.DataFrame(conugeal)
        return conugeal

## Dataset repository
class DatasetRepo:
    """
    Object containing:
      - a Dataset object
      - a report on how the dataset was generated
      - file location
      - a history of the code that generated the current
        dataset and earlier versions
      - a hash of the code that generated the current dataset
      - archived earlier dataset objects

    Method to keep only the most recent version of the
    dataset to free up hard disk space
    """
    def __init__(self,dataset,code,report,filename):
        import hashlib, pickle
        self.dataset = dataset
        self.report = report
        self.currentHash = hashlib.md5(code.encode()).hexdigest()
        self.code = OrderedDict([(self.currentHash,code)])
        self.archive = OrderedDict()
        self.filename = filename
        pickle.dump(self,open(filename,'wb'))

    def update(self,dataset,code,report):
        import hashlib, pickle
        # Put previous object in archive
        self.archive[self.currentHash] = self.dataset

        # Update
        self.dataset = dataset
        self.report = report
        self.currentHash = hashlib.md5(code.encode()).hexdigest()
        self.code[self.currentHash] = code
        pickle.dump(self,open(self.filename,'wb'))

    def wipeArchive(self):
        self.archive = OrderedDict()
        pickle.dump(self,open(self.filename,'wb'))
        
def retrieveSources(dataset_getfunction):
    """
    A dataset_getfunction function that contains 'Source:' lines
    in the docstring, can be decorated with this function.
    If a source is not locally available, it will be downloaded
    and added to the processedDataStorage location.

    A source line has to be formatted accordingly:
    Source: [filename] url

    If filename is not provided, the last part of the url (after last '/')
    is taken as filename.

    Source lines can also be provided as arguments to the FileNotFoundError
    that the dataset_getfunction throws.
    """
    import inspect, requests
    from urllib.request import urlopen, urlretrieve

    def wrapper(*args, **kwargs):
        try:
            return dataset_getfunction(*args, **kwargs)
        except FileNotFoundError as fnf:
            for docline in inspect.getdoc(dataset_getfunction).split('\n') + list(fnf.args):
                if docline.startswith('Source:'):
                    docline = docline.split()
                    if len(docline) == 2:
                        url = docline[1]
                        filename = url[url.rindex('/')+1:]
                    elif len(docline) == 3:
                        url = docline[2]
                        filename = docline[1]
                    if not exists(os.path.join(processedDataStorage,filename)):
                        print('Downloading {}:'.format(filename))
                        if url.startswith('ftp://'):
                            urlretrieve(url,os.path.join(processedDataStorage,filename),
                                        lambda x,y,z: print("\r[{}{}]".format('=' * int(50*x*y/z),
                                                                              ' ' * (50-int(50*x*y/z))),
                                                            end='',flush=True))
                        else:
                            r = requests.get(url,stream=True)
                            total_length = r.headers.get('content-length')
                            with open(processedDataStorage+filename,'wb') as f:
                                if total_length is None: f.write(r.content)
                                else:
                                    dl = 0
                                    total_length = int(total_length)
                                    for data in r.iter_content(chunk_size=4096):
                                        dl += len(data)
                                        f.write(data)
                                        done = int(50 * dl / total_length)
                                        print("\r[{}{}]".format('=' * done, ' ' * (50-done)),end='',flush=True)
            try: return dataset_getfunction(*args, **kwargs)
            except FileNotFoundError:
                print('Either not all source files are documented correctly in docstring,',
                      'or there is a source file unrelated issue')
                raise

    return wrapper

def cacheable(importer,exporter,extension='.cache'):
    """
    Produces decorators with a specified importer and exporter function.
    For example, if a function produces a pandas DataFrame, the importer,
    and exporter could be respectively a lambda for DataFrame.read_csv 
    and DataFrame.to_csv

    extension specifies the file extension used by the caching functions
    """
    def cachedecorator(function):
        def wrapper(*args,cache=True,cache_name='',**kwargs):
            #produce function call hash -> function name, args and kwargs should be merged, stringified and hashed
            import inspect, hashlib, time, datetime
            signature = inspect.signature(function)
            boundArgs = signature.bind(*args,**kwargs)
            boundArgs.apply_defaults()
            callhash = '{}_{}'.format(
                function.__name__ if not cache_name else cache_name,
                hashlib.md5(boundArgs.__repr__().encode()).hexdigest()
            )
            cachedir = config['LSD']['cachedir']
            cachefile = os.path.join(cachedir,'{}{}'.format(callhash,extension))
            # If cache, check if cache exists and how long it is allowed to exist in config
            timeMap = {'h': 'hours', 'd': 'days', 'w': 'weeks'}
            cacheAllowedTime = config['LSD']['cachetime']
            cacheAllowedTime = datetime.timedelta(
                **{timeMap[t]:int(cacheAllowedTime[:-1]) for t in timeMap if cacheAllowedTime.endswith(t)}
            )
            tTBM = time.time() - cacheAllowedTime.total_seconds() # time To Be Modified
            #time.strftime('%c',time.gmtime(tTBM))
            if cache and os.path.exists(cachefile) and os.path.getmtime(cachefile) > tTBM: #within allowed cache time
                return importer(cachefile)
            elif cache:
                # Check if cachedir exists
                if not os.path.exists(cachedir):
                    raise FileNotFoundError(
                        "LSD cache dir ({}) does not exist. Create, change in config or run function with cache=False".format(cachedir)
                    )
                # Redirect stdout and stderr
                stouterr_redirect = StringIO()
                with redirect_stdout(stouterr_redirect), redirect_stderr(stouterr_redirect):
                    functionData = function(*args,**kwargs)
                stouterr_function = stouterr_redirect.getvalue().strip()
                if stouterr_function: print(stouterr_function)
                exporter(functionData,cachefile,stouterr_function)
                return functionData
            else:
                return function(*args,**kwargs)
            
        return wrapper
    return cachedecorator

#cacheableTable reads from and writes to csv, does not log function output
cacheableTable = cacheable(
    importer = lambda x: pd.read_csv(x,index_col=0),
    exporter = lambda x,y,z: x.to_csv(y),
    extension = '.csv'
)

def storeDatasetLocally(dataset_getfunction):
    """
    Can be used as a decorator for 'get_dataset' functions.
    It will check if a processed dataset is locally available,
    and if so, load that one instead of processing from the source
    files.

    Should only be used for functions that do not process the data
    differently depending on the 'get_dataset' function arguments.
    The wrapper raises a warning if there are arguments to pass to 
    the 'get_dataset' function.

    As opposed to cacheable decorated functions, storeDatasetLocally
    is intended for volatile source code, such as e.g. your own scripts
    for which it is nonetheless useful being able to cache results. 
    """
    import inspect, hashlib, re
    from plumbum import colors
    dependency = re.compile(r'\W*Dependenc(y|ies): (.+)')
    
    def wrapper(*args, verbose=True, **kwargs):
        if args or kwargs:
            import warnings
            warnings.warn(
                'This decorated function is not designed to use with arguments. Be warned!'
            )
        # Check if data was already processed
        ## Prepare hash
        functionSource = inspect.getsource(dataset_getfunction).encode()
        ### Check if dependencies
        docstr = inspect.getdoc(dataset_getfunction)
        if docstr:
            dependencies = [e for d in (d.groups()[1].split()
                                        for d in (dependency.search(l)
                                                  for l in docstr.split('\n')) if d)
                            for e in d]
            for d in dependencies:
                d = d.split('.')
                d = getattr(globals()[d[0]],d[-1],globals()[d[0]]) #hack to get d with globals and getattr
                try:
                    functionSource+=inspect.getsource(d).encode()
                except TypeError:
                    functionSource+=pickle.dumps(d)
        hashvalue = hashlib.md5(functionSource).hexdigest()
        datastorage = '{}{}.pickle'.format(processedDataStorage,
                                              dataset_getfunction.__name__.replace('get_','')
        )
        
        if exists(datastorage):
            with open(datastorage,'rb') as openedDatastorage:
                datasetrepo = pickle.load(openedDatastorage)
            if datasetrepo.currentHash == hashvalue:
                print(datasetrepo.report)
                if verbose:
                    print(colors.green & colors.bold | 'Repo size {:.1f}MB, archive contains {} other versions'.format(
                        os.stat(datastorage).st_size/1024**2,
                        len(datasetrepo.archive)
                    ))
                return datasetrepo.dataset
            else:
                print(colors.cyan | 'Dataset content out of date, updating:')
                updateRepo = True
        else:
            print(colors.cyan | 'Dataset not locally available, generating:')
            updateRepo = False

        # Redirect stdout and stderr
        stouterr_redirect = StringIO()
        with redirect_stdout(stouterr_redirect), redirect_stderr(stouterr_redirect):
            start = time.process_time()
            # Run the function that generates the dataset
            dataset = dataset_getfunction(*args, **kwargs)
            duration = time.process_time() - start
            print('Dataset',datastorage,'generated',time.strftime('%c'))
            print('Processing time',time.strftime('%H:%M:%S', time.gmtime(duration)))
            
        print(stouterr_redirect.getvalue())

        if updateRepo:
            datasetrepo.update(dataset,functionSource.decode(),stouterr_redirect.getvalue().strip())
        else:
            try: DatasetRepo(dataset,functionSource.decode(),stouterr_redirect.getvalue().strip(),datastorage)
            except FileNotFoundError:
                print('Not possible to store dataset locally. Create',processedDataStorage,
                      'if you want to avoid reprocessing dataset on every call.')
            
        return dataset

    return wrapper

## Datasets
### References/annotations
from .dealer.ensembl import get_ensembl, get_ensemblGeneannot

@retrieveSources
def get_liftover(frm=19,to=38):
    """
    Info: http://hgdownload.cse.ucsc.edu/downloads.html
    """
    from pyliftover import LiftOver
    liftoverfile = 'hg{}ToHg{}.over.chain.gz'.format(frm,to)
    try: return LiftOver(processedDataStorage+liftoverfile)
    except FileNotFoundError:
        raise FileNotFoundError('Source: http://hgdownload.cse.ucsc.edu/gbdb/hg{}/liftOver/{}'.format(frm,liftoverfile))

def get_lift19to38():
    """
    Source: http://hgdownload.cse.ucsc.edu/gbdb/hg19/liftOver/hg19ToHg38.over.chain.gz
    """
    return get_liftover(frm=19,to=38)

@storeDatasetLocally
def get_proteinNetworks():
    """
    Source: https://thebiogrid.org/downloads/archives/Release%20Archive/BIOGRID-3.4.147/BIOGRID-ALL-3.4.147.tab2.zip
    Source: http://string-db.org/download/protein.links.v10/9606.protein.links.v10.txt.gz
    Source: http://string-db.org/mapping_files/entrez_mappings/entrez_gene_id.vs.string.v10.28042015.tsv
    """
    import networkx as nx
    from bidali.LSD.dealer.external import entrez
    
    #Biogrid
    with ZipFile(datadir+'ProteinNetworks/BIOGRID-ALL-3.4.147.tab2.zip') as biogridzip:
        ds = pd.read_table(TextIOWrapper(biogridzip.open('BIOGRID-ALL-3.4.147.tab2.txt','r')),low_memory=False)
    ds = ds[ds['Organism Interactor A'] == 9606]
    Gbio = nx.Graph()
    ds.T.apply(lambda x: Gbio.add_edge(x['Official Symbol Interactor A'],x['Official Symbol Interactor B']))

    #String-DB
    stringdb = pd.read_table(gzip.open(datadir+'ProteinNetworks/9606.protein.links.v10.txt.gz','rt'),sep=' ')
    stringids = pd.read_table(datadir+'ProteinNetworks/entrez_gene_id.vs.string.v10.28042015.tsv',index_col='STRING_Locus_ID')
    entrez = entrez.get_refseq()
    stringids = stringids[stringids['#Entrez_Gene_ID'].isin(entrez.index)]
    stringdb = stringdb[stringdb.combined_score > 400] #study stringdb.combined_score.hist(bins='auto') to set threshold
    stringdb = stringdb[stringdb.protein1.isin(stringids.index) & stringdb.protein2.isin(stringids.index)]
    stringdb.protein1 = stringdb.protein1.apply(lambda x: stringids.loc[x]['#Entrez_Gene_ID'])
    stringdb.protein2 = stringdb.protein2.apply(lambda x: stringids.loc[x]['#Entrez_Gene_ID'])
    stringdb.protein1 = stringdb.protein1.apply(lambda x: entrez.loc[x].Symbol)
    stringdb.protein2 = stringdb.protein2.apply(lambda x: entrez.loc[x].Symbol)
    Gstring = nx.Graph()
    stringdb.T.apply(lambda x: Gstring.add_edge(x.protein1,x.protein2))
    
    return Dataset(biogridnx = Gbio, biogrid = ds,
                   stringnx = Gstring, string = stringdb)
    
@storeDatasetLocally
def get_centromeres():
    """
    Source: from R bioconductor GWASTools: data(centromeres.hg38)
    """
    from bidali.seqanalysis import loadHumanGenome

    # Centromere positions
    centromereshg38="""1      1 122026460  125184587
2      2  92188146   94090557
3      3  90772459   93655574
4      4  49708101   51743951
5      5  46485901   50059807
6      6  58553889   59829934
7      7  58169654   60828234
8      8  44033745   45877265
9      9  43236168   45518558
10    10  39686683   41593521
11    11  51078349   54425074
12    12  34769408   37185252
13    13  16000001   18051248
14    14  16000001   18173523
15    15  17000001   19725254
16    16  36311159   38280682
17    17  22813680   26885980
18    18  15460900   20861206
19    19  24498981   27190874
20    20  26436233   30038348
21    21  10864561   12915808
22    22  12954789   15054318
X      X  58605580   62412542
Y      Y  10316945   10544039"""

    centromereshg38 = pd.DataFrame([c.split()[-3:] for c in centromereshg38.split('\n')],
                                   columns= "chrom left_base right_base".split())
    centromereshg38.index = centromereshg38.chrom.apply(lambda x: 'chr'+x)
    centromereshg38['left_base'] = centromereshg38.pop('left_base').apply(int)
    centromereshg38['right_base']=centromereshg38.pop('right_base').apply(int)

    genome = loadHumanGenome()
    centromereshg38['len'] = centromereshg38.apply(lambda x: len(genome.chromosomes[x.name]),axis=1)
    centromereshg38['qlen'] = centromereshg38.len - centromereshg38.right_base
    centromereshg38['chr_weight'] = centromereshg38.len/centromereshg38.len.max()
    centromereshg38['q_weight'] = centromereshg38.qlen/centromereshg38[['left_base','qlen']].max().max()
    centromereshg38['p_weight'] = centromereshg38.left_base/centromereshg38[['left_base','qlen']].max().max()
    del genome

    ensembl = get_ensembl()
    ensembl = ensembl[ensembl.chr.isin(centromereshg38.index)]
    ensembl['chrarm'] = ensembl.apply(lambda x: 'p' if x.stop < centromereshg38.loc[x.chr].left_base else
                                  ('q' if x.start > centromereshg38.loc[x.chr].right_base else 'pq'),axis=1)
    centromereshg38['chr_genes'] = ensembl.groupby('chr').size()
    centromereshg38['p_genes'] = ensembl[ensembl.chrarm=='p'].groupby('chr').size()
    centromereshg38['q_genes'] = ensembl[ensembl.chrarm=='q'].groupby('chr').size()
    centromereshg38['chr_gweight'] = centromereshg38.chr_genes/centromereshg38.chr_genes.max()
    centromereshg38['q_gweight'] = centromereshg38.q_genes/centromereshg38[['p_genes','q_genes']].max().max()
    centromereshg38['p_gweight'] = centromereshg38.p_genes/centromereshg38[['p_genes','q_genes']].max().max()
    del ensembl

    return centromereshg38

#@retrieveSources -> not working login required
def get_msigdb6():
    """
    Source: http://software.broadinstitute.org/gsea/msigdb/download_file.jsp?filePath=/resources/msigdb/5.2/msigdb_v6.0.xml
    """
    import xml.etree.ElementTree as ET
    import pickle
    parser = ET.parse(processedDataStorage+'msigdb_v6.0.xml')
    root = parser.getroot()
    genesetsCollections = {} 
    for geneset in root:
        if (geneset.attrib['CATEGORY_CODE'] == 'ARCHIVED' or
            geneset.attrib['ORGANISM'] != 'Homo sapiens'): continue
        try:
            genesetsCollections[geneset.attrib['CATEGORY_CODE']
            ][geneset.attrib['STANDARD_NAME']] = geneset.attrib['MEMBERS_SYMBOLIZED'].split(',')
        except KeyError as e:
            genesetsCollections[geneset.attrib['CATEGORY_CODE']] = {}
            genesetsCollections[geneset.attrib['CATEGORY_CODE']
            ][geneset.attrib['STANDARD_NAME']] = geneset.attrib['MEMBERS_SYMBOLIZED'].split(',')
    genesetsCollections = {'version':'6','MSigDB':genesetsCollections}        
    print('MSigDB {}'.format(genesetsCollections['version']))
    mdb = genesetsCollections['MSigDB']
    return mdb

